from .exceptions import (
    LoaderFormatterException,
)
from .abc import (
    LoaderFormatter,
)
from .hashcode import (
    HashCodeLoaderFormatter,
)
from .cordeau_laporte import (
    CordeauLaporteLoaderFormatter,
)



from .abc import (
    Algorithm,
)
from .naive import (
    NaiveAlgorithm,
)
from .heuristics import (
    InsertionAlgorithm,
    LocalSearchAlgorithm,
)
from .metaheuristics import (
    GraspAlgorithm,
    IterativeAlgorithm,
)
from .exacts import (
    MilpAlgorithm,
)
from .utils import (
    Crosser,
    StatelessCrosser,
    BestStatelessCrosser,
    OrderedCrosser,
    RandomizedCrosser,
    Breeder,
    FlipBreeder,
    Conjecturer,
    SamplingConjecturer,
    IntensiveConjecturer,
)



from __future__ import annotations

import logging
from collections import (
    defaultdict,
)
from itertools import (
    product,
)
from operator import (
    itemgetter,
)
from typing import (
    TYPE_CHECKING,
)

from cached_property import (
    cached_property,
)
import pulp as lp

from ....models import (
    Stop,
    Route,
    PlannedTrip,
)
from .abc import (
    Model,
)

if TYPE_CHECKING:
    from typing import (
        List,
        Set,
        Tuple,
        Iterable,
        Optional,
        Dict,
    )
    from ....models import (
        Trip,
        Vehicle,
        Position,
    )

logger = logging.getLogger(__name__)

BIG = 10000


class ThreeIndexModel(Model):

    def __init__(self, solver: lp.LpSolver = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.solver = solver

    @cached_property
    def problem(self) -> lp.LpProblem:
        problem = lp.LpProblem("3-idx_dial-a-ride", lp.LpMinimize)
        problem.objective = self.objective
        problem.extend(self.constraints)
        return problem

    @cached_property
    def objective(self) -> Optional[lp.LpConstraintVar]:
        return lp.lpSum(
            self.x[k][i][j] * self.costs[i][j]
            for k, i, j in product(self.routes_indexer, self.positions_indexer, self.positions_indexer)
        )

    @cached_property
    def constraints(self) -> List[lp.LpConstraint]:
        constraints: List[lp.LpConstraint] = sum([
            self.uniqueness_constraints,
            self.connectivity_constraints,
            self.time_constraints,
            self.feasibility_constraints,
        ], [])

        logger.info(f'Built "{len(constraints)}" constraints.')
        return constraints

    @cached_property
    def x(self) -> List[List[List[lp.LpVariable]]]:
        x = list()
        for k in self.routes_indexer:
            x_k = list()
            for i in self.positions_indexer:
                x_ki = list()
                for j in self.positions_indexer:
                    x_kij = lp.LpVariable(f'x_{k}_{i}_{j}', cat=lp.LpBinary)
                    x_ki.append(x_kij)
                x_k.append(x_ki)
            x.append(x_k)
        return x

    @cached_property
    def u(self) -> List[List[lp.LpVariable]]:
        u = list()
        for k in self.routes_indexer:
            u_k = list()
            for i in self.positions_indexer:
                u_ki = lp.LpVariable(f'u_{k}_{i}', lowBound=0.0)
                u_k.append(u_ki)
            u.append(u_k)
        return u

    @cached_property
    def w(self) -> List[List[lp.LpVariable]]:
        w = list()
        for k in self.routes_indexer:
            w_k = list()
            for i in self.positions_indexer:
                w_ki = lp.LpVariable(f'w_{k}_{i}', lowBound=0.0)
                w_k.append(w_ki)
            w.append(w_k)
        return w

    @cached_property
    def vehicles(self) -> Tuple[Vehicle, ...]:
        return tuple(self.fleet.vehicles)

    @cached_property
    def trips(self) -> Tuple[Trip, ...]:
        return tuple(self.job.trips)

    @cached_property
    def positions(self) -> Tuple[Position, ...]:
        origins = tuple(trip.origin_position for trip in self.trips)
        destinations = tuple(trip.destination_position for trip in self.trips)
        positions = (
                (self.vehicles[0].origin_position,) +
                origins +
                destinations +
                (self.vehicles[0].destination_position,)
        )

        return positions

    @cached_property
    def costs(self):
        costs = list()
        for origin in self.positions:
            origin_costs = list()
            for destination in self.positions:
                if origin == destination:
                    cost = BIG
                else:
                    cost = origin.distance_to(destination)
                origin_costs.append(cost)
            costs.append(origin_costs)
        return costs

    @property
    def pickups_indexer(self) -> Iterable[int]:
        return range(1, self.n + 1)

    @property
    def nodes_indexer(self) -> Iterable[int]:
        return range(1, self.n * 2 + 1)

    @property
    def routes_indexer(self) -> Iterable[int]:
        return range(len(self.vehicles))

    @property
    def positions_indexer(self) -> Iterable[int]:
        return range(len(self.positions))

    @property
    def n(self) -> int:
        return len(self.trips)

    def trip_by_position_idx(self, idx: int) -> Optional[Trip]:
        if idx in (0, len(self.positions) - 1):
            return None
        return self.trips[(idx % self.n) - 1]

    def time_window_by_position_idx(self, idx: int) -> Tuple[float, float]:
        position = self.positions[idx]
        trip = self.trip_by_position_idx(idx)
        if trip is None:
            earliest, latest = 0, 1440
        elif position == trip.origin_position:
            earliest, latest = trip.origin_earliest, trip.origin_latest
        elif position == trip.destination_position:
            earliest, latest = trip.destination_earliest, trip.destination_latest
        else:
            raise Exception(f'There was a problem related with earliest, latest indices.')
        return earliest, latest

    def capacity_by_position_idx(self, idx: int) -> float:
        trip = self.trip_by_position_idx(idx)
        if trip is None:
            return 0

        capacity = trip.capacity
        if not idx < len(self.positions) / 2:
            capacity *= -1

        return capacity

    def load_time_by_position_idx(self, idx: int) -> float:
        trip = self.trip_by_position_idx(idx)
        if trip is None:
            return 0
        return trip.origin_duration

    def timeout_by_position_idx(self, idx: int) -> float:
        trip = self.trip_by_position_idx(idx)
        if trip is None:
            return 0
        return trip.timeout

    @cached_property
    def uniqueness_constraints(self) -> List[lp.LpConstraint]:
        constraints = list()

        for i in self.pickups_indexer:
            lhs = lp.lpSum(
                self.x[k][i][j]
                for j, k in product(self.positions_indexer, self.routes_indexer)
            )
            constraints.append(lhs == 1)

            for k in self.routes_indexer:
                lhs = (
                        lp.lpSum(self.x[k][i][j] for j in self.positions_indexer) -
                        lp.lpSum(self.x[k][self.n + i][j] for j in self.positions_indexer)
                )
                constraints.append(lhs == 0)

        return constraints

    @cached_property
    def connectivity_constraints(self) -> List[lp.LpConstraint]:
        constraints = list()

        for k in self.routes_indexer:
            lhs = lp.lpSum(self.x[k][0][j] for j in self.positions_indexer)
            constraints.append(lhs == 1)

            for i in self.nodes_indexer:
                lhs = (
                        lp.lpSum(self.x[k][j][i] for j in self.positions_indexer) -
                        lp.lpSum(self.x[k][i][j] for j in self.positions_indexer)
                )
                constraints.append(lhs == 0)

            lhs = lp.lpSum(self.x[k][i][-1] for i in self.positions_indexer)
            constraints.append(lhs == 1)

        return constraints

    @cached_property
    def time_constraints(self) -> List[lp.LpConstraint]:
        constraints = list()

        for k in self.routes_indexer:
            for i, j in product(self.positions_indexer, self.positions_indexer):
                load_time_i = self.load_time_by_position_idx(i)
                travel_time = self.positions[i].time_to(self.positions[j])

                earliest_i, latest_i = self.time_window_by_position_idx(i)
                earliest_j, latest_j = self.time_window_by_position_idx(j)

                cons = max(0, latest_i + load_time_i + travel_time - earliest_j)

                constraints.append(
                    self.u[k][j] >= self.u[k][i] + load_time_i + travel_time - cons * (1 - self.x[k][i][j]),
                )

                capacity_i = self.capacity_by_position_idx(i)
                capacity_j = self.capacity_by_position_idx(j)

                cons = self.vehicles[k].capacity + min(0.0, capacity_i)

                constraints.append(
                    self.w[k][j] >= self.w[k][i] + capacity_j - cons * (1 - self.x[k][i][j]),
                )

        return constraints

    @cached_property
    def feasibility_constraints(self) -> List[lp.LpConstraint]:
        constraints = list()
        for k in self.routes_indexer:
            constraint = self.u[k][-1] - self.u[k][0] <= self.vehicles[k].timeout
            constraints.append(constraint)

            for i in self.positions_indexer:
                earliest, latest = self.time_window_by_position_idx(i)

                constraints.extend([
                    earliest <= self.u[k][i],
                    self.u[k][i] <= latest,
                ])

            for i in self.pickups_indexer:
                load_time = self.load_time_by_position_idx(i)
                timeout = self.timeout_by_position_idx(i)
                travel_time = self.positions[i].time_to(self.positions[i + self.n])

                constraints.extend([
                    travel_time <= self.u[k][i + self.n] - (self.u[k][i] + load_time),
                    self.u[k][i + self.n] - (self.u[k][i] + load_time) <= timeout,
                ])

            for i in self.positions_indexer:
                capacity = self.capacity_by_position_idx(i)

                constraints.extend([
                    max(0.0, capacity) <= self.w[k][i],
                    self.w[k][i] <= self.vehicles[k].capacity + min(0.0, capacity),
                ])

        return constraints

    def solve(self) -> Set[Route]:
        logger.info('Starting to solve...')
        self.problem.solve(self.solver)

        self.validate()
        logger.info(f'Obtained "{lp.value(self.objective)}" reaching "{lp.LpStatus[self.problem.status]}".')
        return self._solution_to_routes()

    def print_solution(self):
        print('X:')
        for k in self.routes_indexer:
            print(f'Vehicle {k}-th.')
            print(f'   {" ".join(map(lambda num: f"{num:02d}", self.positions_indexer))}')
            for i in self.positions_indexer:
                print(f'{i:02d}', end=' ')
                for j in self.positions_indexer:
                    print(f'{int(self.x[k][i][j].varValue):2d}', end=' ')
                print()
            print()

        print('U:')
        for k in self.routes_indexer:
            print(f'Vehicle {k}-th.')
            for i in self.positions_indexer:
                print(f'{self.u[k][i].varValue:4.01f}', end=' ')
            print()

        print('W:')
        for k in self.routes_indexer:
            print(f'Vehicle {k}-th.')
            for i in self.positions_indexer:
                print(f'{self.w[k][i].varValue:4.01f}', end=' ')
            print()

    def validate(self):
        for k in self.routes_indexer:
            for i in self.positions_indexer:
                logger.info(f'Obtained "u[{k}][{i}]={self.u[k][i].varValue}".')
                assert self.u[k][i].varValue >= 0.0

                logger.info(f'Obtained "w[{k}][{i}]={self.w[k][i].varValue}".')
                assert self.w[k][i].varValue >= 0.0

                for j in self.positions_indexer:
                    logger.info(f'Obtained "x[{k}][{i}][{j}]={self.x[k][i][j].varValue}".')
                    assert min(abs(self.x[k][i][j].varValue), abs(self.x[k][i][j].varValue - 1)) <= 0.05

    def _solution_to_routes(self):
        logger.info(f'Casting solution to a set of routes...')
        routes = set()
        for k in self.routes_indexer:
            vehicle = self.vehicles[k]
            route = Route(vehicle)

            ordered_trip_indexes = [
                idx
                for idx, u_k in sorted(enumerate(u_k.varValue for u_k in self.u[k]), key=itemgetter(1))
            ]

            positions = self._solution_to_positions(k, ordered_trip_indexes)
            stops = self._positions_to_stops(route, positions)
            trips = self._positions_to_trips(positions)
            self._build_planned_trips(vehicle, trips, stops)

            for stop in stops:
                route.insert_stop(stop)
            routes.add(route)
        return routes

    def _positions_to_trips(self, positions) -> List[Trip]:
        trips: List[Trip] = list()
        for position in positions:
            trip = next((trip for trip in self.trips if trip.origin_position == position), None)
            if trip is None:
                continue
            if trip in trips:
                continue
            trips.append(trip)
        return trips

    def _build_planned_trips(self, vehicle: Vehicle, trips: List[Trip], stops: List[Stop]) -> List[PlannedTrip]:
        stop_mapper = self._stop_to_stop_mapper(stops)

        planned_trips = list()
        for trip in trips:
            pickup = stop_mapper[trip.origin_position].pop(0)
            delivery = stop_mapper[trip.destination_position].pop(0)

            planned_trip = PlannedTrip(vehicle, trip, pickup, delivery)
            planned_trips.append(planned_trip)
        return planned_trips

    def _solution_to_positions(self, k, ordered_trip_indexes) -> List[Position]:
        positions = list()
        for i, j in product(ordered_trip_indexes, self.positions_indexer):
            if not int(self.x[k][i][j].varValue) == 1:
                continue
            positions.append(self.positions[i])
        return positions

    def _positions_to_stops(self, route, positions) -> List[Stop]:
        stops = [route.first_stop]
        for position in positions:
            pickup = Stop(route, position, stops[-1])
            stops.append(pickup)
        stops.pop(0)
        return stops

    def _stop_to_stop_mapper(self, stops: List[Stop]) -> Dict[Position, List[Stop]]:
        mapper: Dict[Position, List[Stop]] = defaultdict(list)
        for stop in stops:
            mapper[stop.position].append(stop)
        return mapper



from .abc import (
    Model,
)
from .positions import (
    Position,
    GeometricPosition,
)
from .surfaces import (
    Surface,
    GeometricSurface,
)
from .trips import (
    Trip,
)
from .planned_trips import (
    PlannedTrip,
)
from .stops import (
    Stop,
)
from .services import (
    Service,
)
from .criterions import (
    RouteCriterion,
    ShortestTimeRouteCriterion,
    LongestTimeRouteCriterion,
    LongestUtilTimeRouteCriterion,
    HashCodeRouteCriterion,
)
from .constants import (
    OptimizationDirection,
    DistanceMetric,
    MAX_INT,
    MIN_INT,
    MAX_FLOAT,
    MIN_FLOAT,
)
from .jobs import (
    Job
)
from .vehicles import (
    Vehicle,
    Fleet,
)
from .routes import (
    Route,
)
from .plannings import (
    Planning,
)
from .results import (
    Result,
)
from .objectives import (
    Objective,
    DialARideObjective,
    TaxiSharingObjective,
    HashCodeObjective,
)



from __future__ import annotations

import logging
from random import Random
from typing import TYPE_CHECKING

from ...models import (
    Planning,
    MAX_INT,
)
from ..abc import (
    Algorithm,
)
from ..heuristics import (
    InsertionAlgorithm,
)

if TYPE_CHECKING:
    from typing import (
        Type,
        Optional,
    )
    from ...models import (
        Result,
    )

logger = logging.getLogger(__name__)


class IterativeAlgorithm(Algorithm):

    def __init__(self, episodes: int = 10, algorithm_cls: Type[Algorithm] = None, seed: int = 56, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if algorithm_cls is None:
            algorithm_cls = InsertionAlgorithm
        self.episodes = episodes
        self.algorithm_cls = algorithm_cls
        self.random = Random(seed)

        self.args = args
        self.kwargs = kwargs

    def build_algorithm(self, *args, **kwargs) -> Algorithm:
        args = (*self.args, *args)
        kwargs.update(self.kwargs)

        return self.algorithm_cls(*args, **kwargs)

    def _optimize(self) -> Planning:
        best: Optional[Result] = None
        for i in range(self.episodes):
            seed = self.random.randint(0, MAX_INT)
            current = self.build_algorithm(seed=seed).optimize()
            best = self.objective.best(best, current)

        assert best is not None
        return best.planning



import unittest

import jinete as jit

from tests.utils import (
    generate_one_loader,
)


class TestMilpAlgorithm(unittest.TestCase):
    loader: jit.Loader

    @classmethod
    def setUpClass(cls) -> None:
        loader_cls = generate_one_loader()
        cls.loader = loader_cls()

    @property
    def job(self) -> jit.Job:
        return self.loader.job

    @property
    def fleet(self) -> jit.Fleet:
        return self.loader.fleet

    def test_creation(self):
        algorithm = jit.MilpAlgorithm(
            job=self.job,
            fleet=self.fleet,
        )
        self.assertEqual(algorithm.job, self.job)
        self.assertEqual(algorithm.fleet, self.fleet)

    def test_optimize(self):
        algorithm = jit.MilpAlgorithm(
            job=self.job,
            fleet=self.fleet,
        )
        result = algorithm.optimize()

        # TODO: Properly validate  behaviour of the provided "Result" object.
        self.assertIsNotNone(result)
        self.assertIsInstance(result, jit.Result)
        self.assertEqual(1, result.coverage_rate)

        self.assertEqual(4, result.optimization_function[0])
        self.assertAlmostEqual(-85.4, result.optimization_function[1], delta=0.1)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from random import uniform, randint
from uuid import uuid4

import jinete as jit

if TYPE_CHECKING:
    from typing import (
        Set,
        Type,
        Optional,
    )


def generate_one_surface(*args, **kwargs) -> jit.Surface:
    kwargs['metric'] = jit.DistanceMetric.MANHATTAN
    return jit.GeometricSurface(*args, **kwargs)


def generate_surfaces(n: int, *args, **kwargs) -> Set[jit.Surface]:
    return {
        generate_one_surface(*args, **kwargs) for _ in range(n)
    }


def generate_one_position(x_min: float = -100, x_max: float = 100, y_min: float = -100,
                          y_max: float = 100, surface: jit.Surface = None, *args, **kwargs) -> jit.Position:
    if surface is None:
        surface = generate_one_surface(*args, **kwargs)

    return surface.get_or_create_position([uniform(x_min, x_max), uniform(y_min, y_max)])


def generate_positions(n: int, surface: jit.Surface = None, *args, **kwargs) -> Set[jit.Position]:
    if surface is None:
        surface = generate_one_surface(*args, **kwargs)
    kwargs['surface'] = surface
    return {
        generate_one_position(*args, **kwargs) for _ in range(n)
    }


def generate_one_trip(identifier: str = None,
                      origin_position: jit.Position = None, destination_position: jit.Position = None,
                      earliest: float = None, earliest_min: float = 0, earliest_max: float = 86400,
                      timeout: float = None, timeout_min: float = 1800, timeout_max: float = 7200,
                      load_time: float = None, load_time_min: float = 300, load_time_max: float = 900,
                      capacity: float = None, capacity_min: int = 1, capacity_max: int = 3,
                      *args, **kwargs) -> jit.Trip:
    if identifier is None:
        identifier = f'{randint(1, 999):0d}'
    if origin_position is None:
        origin_position = generate_one_position(*args, **kwargs)
    if destination_position is None:
        destination_position = generate_one_position(*args, **kwargs)
    if earliest is None:
        earliest = uniform(earliest_min, earliest_max)
    if timeout is None:
        timeout = uniform(timeout_min, timeout_max)
    if capacity is None:
        capacity = randint(capacity_min, capacity_max)
    if load_time is None:
        load_time = uniform(load_time_min, load_time_max)

    return jit.Trip(
        identifier,
        origin=jit.Service(
            position=origin_position,
            earliest=earliest,
            latest=earliest + timeout,
            duration=load_time,
        ),
        destination=jit.Service(
            position=destination_position,
        ),
        capacity=capacity,
    )


def generate_trips(n: int, *args, **kwargs) -> Set[jit.Trip]:
    return {
        generate_one_trip(str(i), *args, **kwargs) for i in range(n)
    }


def generate_one_job(trips_count: int = None, trips_count_min: int = 1, trips_count_max: int = 100,
                     objective_cls: Optional[Type[jit.Objective]] = None, *args, **kwargs) -> jit.Job:
    if trips_count is None:
        trips_count = randint(trips_count_min, trips_count_max)
    if objective_cls is None:
        objective_cls = jit.DialARideObjective

    trips = generate_trips(trips_count, *args, **kwargs)
    job = jit.Job(trips, objective_cls)
    return job


def generate_one_planning(routes_count: int = None, routes_count_min: int = 1, routes_count_max: int = 100,
                          *args, **kwargs) -> jit.Planning:
    if routes_count is None:
        routes_count = randint(routes_count_min, routes_count_max)
    trips = generate_routes(routes_count, *args, **kwargs)
    planning = jit.Planning(trips)
    return planning


def generate_jobs(n: int, *args, **kwargs) -> Set[jit.Job]:
    return {
        generate_one_job(*args, **kwargs) for _ in range(n)
    }


def generate_one_planned_trip(feasible: bool, vehicle: jit.Vehicle = None, previous_stop: jit.Stop = None,
                              pickup_stop: jit.Stop = None, delivery_stop: jit.Stop = None,
                              origin_position: jit.Position = None, destination_position: jit.Position = None,
                              *args, **kwargs) -> jit.PlannedTrip:
    if vehicle is None:
        vehicle = generate_one_vehicle()
    if previous_stop is None:
        previous_stop = jit.Stop(vehicle, vehicle.origin_position, previous_stop)

    if feasible:
        down_time = 0.0
        kwargs['earliest'] = 0.0
        kwargs['timeout'] = float('inf')
        kwargs['capacity'] = vehicle.capacity
    else:
        down_time = jit.MAX_FLOAT

    if origin_position is None:
        origin_position = generate_one_position()

    if destination_position is None:
        destination_position = generate_one_position()

    if pickup_stop is None:
        pickup_stop = jit.Stop(vehicle, origin_position, previous_stop)
    else:
        origin_position = pickup_stop.position

    if delivery_stop is None:
        delivery_stop = jit.Stop(vehicle, destination_position, pickup_stop)
    else:
        destination_position = delivery_stop.position

    trip = generate_one_trip(
        origin_position=origin_position,
        destination_position=destination_position,
        *args, **kwargs,
    )

    return jit.PlannedTrip(
        vehicle=vehicle,
        trip=trip,
        pickup=pickup_stop,
        delivery=delivery_stop,
        down_time=down_time,
    )


def generate_one_vehicle(capacity_min: int = 1, capacity_max: int = 3, earliest_min: float = 0,
                         earliest_max: float = 86400,
                         timeout: float = None, timeout_min: float = 14400, timeout_max: float = 28800,
                         idx: int = 0, *args, **kwargs) -> jit.Vehicle:
    # TODO: Increase parameter options.
    capacity = randint(capacity_min, capacity_max)
    position = generate_one_position(*args, **kwargs)
    earliest = uniform(earliest_min, earliest_max)
    if timeout is None:
        timeout = uniform(timeout_min, timeout_max)
    latest = earliest + timeout

    origin = jit.Service(position=position, earliest=earliest, latest=latest)
    return jit.Vehicle(str(idx), origin, capacity=capacity)


def generate_vehicles(n: int, *args, **kwargs) -> Set[jit.Vehicle]:
    vehicles = set()
    for idx in range(n):
        kwargs['idx'] = idx
        vehicle = generate_one_vehicle(*args, **kwargs)
        vehicles.add(vehicle)
    return vehicles


def generate_one_route(feasible: bool = True,
                       planned_trips_count: int = None, planned_trips_min: int = 1, planned_trips_max: int = 20,
                       surface: jit.Surface = None, *args, **kwargs) -> jit.Route:
    if surface is None:
        surface = generate_one_surface(*args, **kwargs)
    if feasible:
        kwargs['timeout'] = float('inf')

    vehicle = generate_one_vehicle(surface=surface, *args, **kwargs)
    route = jit.Route(vehicle)

    if planned_trips_count is None:
        planned_trips_count = randint(planned_trips_min, planned_trips_max)

    for i in range(planned_trips_count):
        planned_trip = generate_one_planned_trip(
            feasible=feasible,
            vehicle=vehicle,
            previous_stop=route.current_stop,
            surface=surface,
        )

        route.append_planned_trip(planned_trip)
    return route


def generate_routes(n: int, *args, **kwargs) -> Set[jit.Route]:
    return {
        generate_one_route(*args, **kwargs) for _ in range(n)
    }


def generate_one_loader() -> Type[jit.Loader]:
    file_path = Path(__file__).parent / 'res' / 'problem-4.txt'

    class MyLoader(jit.FileLoader):
        def __init__(self):
            super().__init__(
                file_path=file_path,
                formatter_cls=jit.CordeauLaporteLoaderFormatter,
            )

    return MyLoader


def generate_one_result():
    from pathlib import Path
    file_path = Path(__file__).parents[1] / 'res' / 'datasets' / 'hashcode' / 'a_example.in'

    class MyLoader(jit.FileLoader):
        def __init__(self, *args, **kwargs):
            super().__init__(
                file_path=file_path,
                formatter_cls=jit.HashCodeLoaderFormatter,
                *args, **kwargs,
            )

    class MyAlgorithm(jit.InsertionAlgorithm):
        def __init__(self, *args, **kwargs):
            super().__init__(
                neighborhood_max_size=None,
                criterion_cls=jit.HashCodeRouteCriterion,
                *args, **kwargs,
            )

    dispatcher = jit.StaticDispatcher(
        MyLoader,
        MyAlgorithm,
    )
    result = dispatcher.run()

    return result



import unittest
from copy import deepcopy

import jinete as jit
from tests.utils import (
    generate_one_position,
)


class TestServices(unittest.TestCase):

    def test_construction(self):
        position = generate_one_position()
        service = jit.Service(position)

        self.assertEqual(position, service.position)
        self.assertEqual(0, service.earliest)
        self.assertEqual(jit.MAX_FLOAT, service.latest)
        self.assertEqual(0, service.duration)

    def test_construction_with_earliest(self):
        position = generate_one_position()
        earliest = 1800
        service = jit.Service(position, earliest=earliest)

        self.assertEqual(position, service.position)
        self.assertEqual(earliest, service.earliest)
        self.assertEqual(jit.MAX_FLOAT, service.latest)
        self.assertEqual(0, service.duration)

    def test_construction_with_latest(self):
        position = generate_one_position()
        latest = 3600
        service = jit.Service(position, latest=latest)

        self.assertEqual(position, service.position)
        self.assertEqual(0, service.earliest)
        self.assertEqual(latest, service.latest)
        self.assertEqual(0, service.duration)

    def test_construction_with_duration(self):
        position = generate_one_position()
        duration = 60
        service = jit.Service(position, duration=duration)

        self.assertEqual(position, service.position)
        self.assertEqual(0, service.earliest)
        self.assertEqual(jit.MAX_FLOAT, service.latest)
        self.assertEqual(duration, service.duration)

    def test_equals(self):
        position = generate_one_position()
        earliest = 1800
        latest = 3600
        duration = 60

        one = jit.Service(position, earliest, latest, duration)
        two = jit.Service(position, earliest, latest, duration)
        self.assertEqual(one, two)

    def test_distance_to(self):
        one_position = generate_one_position()
        two_position = generate_one_position()
        one = jit.Service(one_position)
        two = jit.Service(two_position)
        self.assertEqual(one_position.distance_to(two_position), one.distance_to(two))

    def test_time_to(self):
        one_position = generate_one_position()
        two_position = generate_one_position()
        one = jit.Service(one_position)
        two = jit.Service(two_position)
        self.assertEqual(one_position.time_to(two_position), one.time_to(two))

    def test_tuple(self):
        position = generate_one_position()
        earliest = 1800
        latest = 3600
        duration = 60

        service = jit.Service(position, earliest, latest, duration)

        raw = (
            ('position', position),
            ('earliest', earliest),
            ('latest', latest),
            ('duration', duration),
        )
        self.assertEqual(raw, tuple(service))

    def test_dict(self):
        position = generate_one_position()
        earliest = 1800
        latest = 3600
        duration = 60

        service = jit.Service(position, earliest, latest, duration)

        raw = {
            'position': position,
            'earliest': earliest,
            'latest': latest,
            'duration': duration,
        }
        self.assertEqual(raw, dict(service))

    def test_deepcopy(self):
        position = generate_one_position()
        copied_position = deepcopy(position)
        self.assertEqual(position, copied_position)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import logging
import itertools as it
from typing import TYPE_CHECKING

from .abc import (
    Crosser,
)

if TYPE_CHECKING:
    from typing import (
        Optional,
        Dict,
        List,
    )
    from ....models import (
        PlannedTrip,
        Route,
        Vehicle
    )

logger = logging.getLogger(__name__)


class OrderedCrosser(Crosser):
    ranking: Dict[Vehicle, List[Route]]

    def __init__(self, neighborhood_max_size: int = 250, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if neighborhood_max_size is None:
            neighborhood_max_size = len(self.trips)
        else:
            neighborhood_max_size = min(neighborhood_max_size, len(self.trips))

        self.neighborhood_max_size = neighborhood_max_size
        self.ranking = dict()
        self.initialize_ranking()

    def initialize_ranking(self) -> None:
        logger.info("Initializing ranking...")
        for route in self.attractive_routes:
            self.ranking[route.vehicle] = self.create_sub_ranking(route)
            logger.info(f"Added sub ranking! Currently {len(self.ranking)}.")

    def mark_planned_trip_as_done(self, planned_trip: PlannedTrip) -> None:
        super().mark_planned_trip_as_done(planned_trip)
        self.update_ranking(planned_trip)

    def update_ranking(self, planned_trip: PlannedTrip) -> None:
        logger.debug(f'Updating all rankings due to planned trip with "{planned_trip.trip_identifier}" trip...')
        for route in self.attractive_routes:
            self._update_vehicle_ranking(planned_trip, route.vehicle)

    def _update_vehicle_ranking(self, planned_trip: PlannedTrip, vehicle: Vehicle):
        logger.debug(f'Updating ranking for vehicle with "{vehicle.identifier}" identifier...')
        if vehicle is planned_trip.vehicle or vehicle not in self.ranking:
            self.ranking[vehicle] = self.create_sub_ranking(self.routes_container[vehicle])
        else:
            self.ranking[vehicle] = [route for route in self.ranking[vehicle] if planned_trip.trip not in route.trips]

    def create_sub_ranking(self, route: Route) -> List[Route]:
        logger.debug(f'Creating sub_ranking for vehicle "{route.vehicle_identifier}"...')
        pending_trips = it.islice(self.pending_trips, self.neighborhood_max_size)

        raw_sub_ranking = self.conjecturer.compute(route, pending_trips)

        self.criterion.sorted(raw_sub_ranking, inplace=True)
        return raw_sub_ranking

    def get_planned_trip(self) -> Optional[PlannedTrip]:
        if len(self.ranking) == 0:
            return None

        best = None
        for sub_ranking in self.ranking.values():
            if len(sub_ranking) == 0:
                continue
            current = sub_ranking[0]
            if not current.feasible:
                continue
            best = self.criterion.best(best, current)
        return best



import logging

from .abc import (
    Breeder,
)
from ....models import (
    Result,
    Stop,
    Route,
)

logger = logging.getLogger(__name__)


class FlipBreeder(Breeder):

    def _improve(self) -> Result:
        logger.info(f'Starting to improve "Result" with "{self.__class__.__name__}"...')
        for idx, route in enumerate(self.routes):
            cost = self.objective.optimization_function(route)

            for i in range(1, len(route.stops) - 1):
                j = i + 1
                k = i + 2
                first = route.stops[i]
                second = route.stops[j]
                third = route.stops[k] if k < len(route.stops) else None

                if not set(first.pickups).isdisjoint(second.deliveries):
                    continue
                self.flip(route, first, second, third)

                if not route.feasible or cost == self.objective.best(cost, route):
                    self.flip(route, second, first, third)
                    continue

                cost = self.objective.optimization_function(route)
                logger.info(f'Flipped "{i}"-th and "{j}"-th stops from "{route}".')
        return self.result

    def flip(self, route: Route, previous: Stop, other: Stop, following: Stop = None) -> None:
        assert following is None or following.previous == other
        assert other.previous == previous

        self_index = route.stops.index(other)
        other_index = route.stops.index(previous)
        route.stops[self_index], route.stops[other_index] = route.stops[other_index], route.stops[self_index]

        if following is not None:
            following.previous = previous
        other.previous = previous.previous
        previous.previous = other

        for stop in route.stops[self_index:]:
            stop.flush()



from __future__ import annotations

from .abc import Dispatcher

from ..models import (
    Result,
)


class StaticDispatcher(Dispatcher):

    def run(self) -> Result:
        loader = self.loader_cls()

        job = loader.job
        fleet = loader.fleet
        algorithm = self.algorithm_cls(fleet=fleet, job=job)

        result = algorithm.optimize()

        if self.storer_cls is not None:
            storer = self.storer_cls(result=result)
            storer.store()

        return result



from .crossers import (
    Crosser,
    StatelessCrosser,
    BestStatelessCrosser,
    OrderedCrosser,
    RandomizedCrosser,
)
from .models import (
    Model,
    ThreeIndexModel,
)
from .breeders import (
    Breeder,
    FlipBreeder,
)
from .conjecturer import (
    Conjecturer,
    SamplingConjecturer,
    IntensiveConjecturer,
)



import unittest

import jinete as jit
from tests.utils import (
    generate_one_planned_trip,
    generate_one_route,
)


class TestJineteException(unittest.TestCase):

    def test_creation(self):
        message = 'This is a test exception'
        exc = jit.JineteException(message)

        self.assertIsInstance(exc, jit.JineteException)
        self.assertEqual(exc.message, message)


class TestStopPlannedTripIterationException(unittest.TestCase):

    def test_creation(self):
        exc = jit.StopPlannedTripIterationException()

        self.assertIsInstance(exc, jit.JineteException)
        self.assertEqual(exc.message, 'There are no more Planned Trips to iterate over them.')


class TestNonFeasiblePlannedTripException(unittest.TestCase):

    def test_creation(self):
        planned_trip = generate_one_planned_trip(False)
        exc = jit.NonFeasiblePlannedTripException(planned_trip)

        self.assertIsInstance(exc, jit.JineteException)
        self.assertEqual(exc.planned_trip, planned_trip)
        self.assertEqual(exc.message, f'Planned Trip "{planned_trip}" is not feasible.')


class TestNonFeasibleRouteException(unittest.TestCase):

    def test_creation(self):
        route = generate_one_route(False)
        exc = jit.NonFeasibleRouteException(route)

        self.assertIsInstance(exc, jit.JineteException)
        self.assertEqual(exc.route, route)
        self.assertEqual(exc.message, f'Route "{route}" is not feasible.')


if __name__ == '__main__':
    unittest.main()



import unittest
from typing import List

import jinete as jit

from tests.utils import (
    generate_vehicles,
    generate_trips,
    generate_one_vehicle,
    generate_one_position,
)


class TestFlipBreeder(unittest.TestCase):
    vehicle: jit.Vehicle
    position: jit.Position
    stops: List[jit.Stop]

    def setUp(self) -> None:
        self.vehicle = generate_one_vehicle()
        self.position = generate_one_position()
        stop0 = jit.Stop(self.vehicle, self.position, None)

        stop1 = jit.Stop(self.vehicle, generate_one_position(), stop0)
        stop0.following = stop1

        stop2 = jit.Stop(self.vehicle, generate_one_position(), stop1)
        stop1.following = stop2

        stop3 = jit.Stop(self.vehicle, generate_one_position(), stop2)
        stop2.following = stop3

        stop4 = jit.Stop(self.vehicle, generate_one_position(), stop3)
        stop3.following = stop4

        stops = [stop0, stop1, stop2, stop3, stop4]
        self.stops = stops

        self.route = jit.Route(self.vehicle, self.stops)
        self.planning = jit.Planning({self.route})

        self.job = jit.Job(generate_trips(10), objective_cls=jit.DialARideObjective)
        self.fleet = jit.Fleet(generate_vehicles(10))
        self.result = jit.Result(self.fleet, self.job, jit.Algorithm, self.planning, computation_time=float(0.0))

    def test_creation(self):
        breeder = jit.FlipBreeder(self.result)

        self.assertNotEqual(self.planning, breeder.planning)
        self.assertEqual(1, len(breeder.routes))
        self.assertEqual(tuple(self.route.positions), tuple(next(iter(breeder.routes)).positions))

    def test_flip(self):
        breeder = jit.FlipBreeder(self.result)
        stops = self.route.stops

        breeder.flip(self.route, stops[1], stops[2], stops[3])

        self.assertEqual(stops[0].position, self.stops[0].position)
        self.assertEqual(stops[1].position, self.stops[2].position)
        self.assertEqual(stops[2].position, self.stops[1].position)
        self.assertEqual(stops[3].position, self.stops[3].position)

        for first, second in zip(stops[:-1], stops[1:]):
            self.assertEqual(second.previous, first)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from .stateless import (
    StatelessCrosser,
)

if TYPE_CHECKING:
    from typing import (
        Optional,
    )
    from ....models import (
        PlannedTrip,
    )

logger = logging.getLogger(__name__)


class BestStatelessCrosser(StatelessCrosser):

    def get_planned_trip(self) -> Optional[PlannedTrip]:
        best = self.criterion.best(*self.iterator)
        return best



from __future__ import annotations

import logging
from abc import (
    ABC,
    abstractmethod,
)
from copy import deepcopy
from typing import (
    TYPE_CHECKING,
)
from cached_property import (
    cached_property,
)

from ....exceptions import (
    StopPlannedTripIterationException,
)
from ....models import (
    Route,
    ShortestTimeRouteCriterion,
)
from ..conjecturer import (
    Conjecturer,
)

if TYPE_CHECKING:
    from typing import (
        Set,
        Optional,
        Type,
        Dict,
    )
    from ....models import (
        PlannedTrip,
        Vehicle,
        Fleet,
        Trip,
        Job,
        RouteCriterion,
    )

logger = logging.getLogger(__name__)


class Crosser(ABC):
    fleet: Fleet
    job: Job
    criterion_cls: Type[RouteCriterion]
    routes_container: Dict[Vehicle, Route]

    def __init__(self, fleet: Fleet, job: Job, conjecturer_cls: Type[Conjecturer] = None,
                 criterion_cls: Type[RouteCriterion] = None, routes: Set[Route] = None, *args, **kwargs):
        if conjecturer_cls is None:
            conjecturer_cls = Conjecturer
        if criterion_cls is None:
            criterion_cls = ShortestTimeRouteCriterion

        pending_trips = set(job.trips)
        if routes is None:
            routes = set(Route(vehicle) for vehicle in fleet.vehicles)
        else:
            routes = deepcopy(routes)
            for route in routes:
                pending_trips -= set(route.trips)

        self.fleet = fleet
        self.job = job
        self.routes_container = {
            route.vehicle: route
            for route in routes
        }
        self._attractive_routes = None
        self.pending_trips = pending_trips

        self.conjecturer_cls = conjecturer_cls
        self.criterion_cls = criterion_cls

        self.args = args
        self.kwargs = kwargs

    @property
    def routes(self) -> Set[Route]:
        return set(self.routes_container.values())

    @property
    def attractive_routes(self) -> Set[Route]:
        if self._attractive_routes is None:
            self._attractive_routes = set(route for route in self.routes if any(route.planned_trips))

        if not any(any(route.planned_trips) for route in self._attractive_routes):
            empty_route = next((route for route in self.routes if not any(route.planned_trips)), None)
            if empty_route is not None:
                self._attractive_routes.add(empty_route)
        return self._attractive_routes

    def set_route(self, route: Route) -> None:
        vehicle = route.vehicle
        logger.debug(f'Updating route for vehicle with "{vehicle.identifier}" identifier...')
        old_trips = set(self.routes_container[vehicle].trips)
        self.routes_container[vehicle] = route
        for planned_trip in route.planned_trips:
            if planned_trip.trip in old_trips:
                continue
            self.mark_planned_trip_as_done(planned_trip)

    @cached_property
    def conjecturer(self) -> Conjecturer:
        return self.conjecturer_cls(*self.args, **self.kwargs)

    @cached_property
    def criterion(self) -> RouteCriterion:
        return self.criterion_cls(*self.args, **self.kwargs)

    @property
    def vehicles(self) -> Set[Vehicle]:
        return self.fleet.vehicles

    @property
    def trips(self) -> Set[Trip]:
        return self.job.trips

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            planned_trip = self.get_planned_trip()
            if planned_trip is None:
                raise StopPlannedTripIterationException()
            logger.debug(f'Yielding {planned_trip}...')
            return planned_trip

    @abstractmethod
    def get_planned_trip(self) -> Optional[PlannedTrip]:
        pass

    def mark_planned_trip_as_done(self, planned_trip: PlannedTrip) -> None:
        logger.info(f'Marking trip with "{planned_trip.trip_identifier}" identifier as done '
                    f'over vehicle with "{planned_trip.vehicle_identifier}" identifier...')
        self.mark_trip_as_done(planned_trip.trip)

    def mark_trip_as_done(self, trip: Trip) -> None:
        logger.debug(f'Marked trip with "{trip.identifier}" identifier as done.')
        self.pending_trips.remove(trip)

    def mark_planned_trip_as_undone(self, planned_trip: PlannedTrip) -> None:
        self.mark_trip_as_undone(planned_trip.trip)

    def mark_trip_as_undone(self, trip: Trip) -> None:
        logger.debug(f'Marked trip with "{trip.identifier}" identifier as undone.')
        self.pending_trips.add(trip)

    @property
    def done_trips(self) -> Set[Trip]:
        return self.trips - self.pending_trips

    @property
    def completed(self) -> bool:
        return len(self.pending_trips) == 0



import unittest

import jinete as jit

from tests.utils import (
    generate_vehicles,
    generate_trips,
)


class TestStatelessCrosser(unittest.TestCase):

    def test_creation(self):
        job = jit.Job(generate_trips(10), objective_cls=jit.DialARideObjective)
        fleet = jit.Fleet(generate_vehicles(10))
        dispatcher = jit.StatelessCrosser(
            job=job,
            fleet=fleet,
        )
        self.assertEqual(job, dispatcher.job)
        self.assertEqual(fleet, dispatcher.fleet)
        self.assertEqual(len(job.trips), len(list(dispatcher.iterator)))

    def test_flush(self):
        job = jit.Job(generate_trips(10), objective_cls=jit.DialARideObjective)
        fleet = jit.Fleet(generate_vehicles(10))
        dispatcher = jit.StatelessCrosser(
            job=job,
            fleet=fleet,
        )
        self.assertNotIn('iterator', dispatcher.__dict__)
        list(dispatcher.iterator)
        self.assertIn('iterator', dispatcher.__dict__)
        dispatcher.flush()
        self.assertNotIn('iterator', dispatcher.__dict__)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
)
from .constants import (
    MAX_FLOAT,
)
from .abc import (
    Model,
)

if TYPE_CHECKING:
    from typing import (
        Dict,
        Any,
        Generator,
        Tuple,
    )
    from .positions import (
        Position,
    )

logger = logging.getLogger(__name__)


class Service(Model):
    __slots__ = (
        'position',
        'earliest',
        'latest',
        'duration',
    )
    position: Position
    earliest: float
    latest: float
    duration: float

    def __init__(self, position: Position, earliest: float = 0.0, latest: float = MAX_FLOAT, duration: float = 0.0):
        self.position = position
        self.earliest = earliest
        self.latest = latest
        self.duration = duration

    def __deepcopy__(self, memo: Dict[int, Any]) -> Service:
        return self

    def __eq__(self, other: Service) -> bool:
        return hash(self) == hash(other)

    def __hash__(self):
        return hash(tuple(self))

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        yield from (
            ('position', self.position),
            ('earliest', self.earliest),
            ('latest', self.latest),
            ('duration', self.duration),
        )

    def distance_to(self, other: Service) -> float:
        return self.position.distance_to(other.position)

    def time_to(self, other: Service, *args, **kwargs) -> float:
        return self.position.time_to(other.position, *args, **kwargs)



from __future__ import annotations

import logging
from random import Random
from typing import TYPE_CHECKING

from ...models import (
    Planning,
    MAX_INT,
)
from ..abc import (
    Algorithm,
)
from ..heuristics import (
    LocalSearchAlgorithm,
)
from .iterative import (
    IterativeAlgorithm,
)

if TYPE_CHECKING:
    pass

logger = logging.getLogger(__name__)


class GraspAlgorithm(Algorithm):

    def __init__(self, first_solution_episodes: int = 1, no_improvement_threshold: int = 10,
                 seed: int = 56, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.first_solution_episodes = first_solution_episodes
        self.no_improvement_threshold = no_improvement_threshold
        self.random = Random(seed)

        self.args = args
        self.kwargs = kwargs

    def build_first_solution_algorithm(self, *args, **kwargs) -> Algorithm:
        args = (*self.args, *args)
        kwargs.update(self.kwargs)

        kwargs['episodes'] = self.first_solution_episodes
        kwargs['seed'] = self.random.randint(0, MAX_INT)

        return IterativeAlgorithm(*args, **kwargs)

    def build_local_search_algorithm(self, *args, **kwargs) -> Algorithm:
        args = (*self.args, *args)
        kwargs.update(self.kwargs)

        kwargs['seed'] = self.random.randint(0, MAX_INT),

        return LocalSearchAlgorithm(*args, **kwargs)

    def _optimize(self) -> Planning:
        iterative = self.build_first_solution_algorithm()
        best = iterative.optimize()

        no_improvement_count = 0
        while no_improvement_count < self.no_improvement_threshold:
            no_improvement_count += 1

            current = self.build_local_search_algorithm(initial=best).optimize()
            current = self.build_first_solution_algorithm(routes=current.routes).optimize()
            best = self.objective.best(best, current)

            if best == current:
                no_improvement_count = 0

        return best.planning



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from .constants import (
    OptimizationDirection,
    MIN_FLOAT,
    MAX_FLOAT,
)

if TYPE_CHECKING:
    from typing import (
        Optional,
        List,
        Iterable,
    )
    from .routes import (
        Route,
    )

logger = logging.getLogger(__name__)


class RouteCriterion(ABC):
    def __init__(self, name: str, direction: OptimizationDirection, *args, **kwargs):
        self.name = name
        self.direction = direction

    @abstractmethod
    def scoring(self, route: Route) -> float:
        pass

    def best(self, *args: Optional[Route]) -> Route:
        return self.direction(
            (arg for arg in args if arg is not None),
            key=self.scoring,
            default=None,
        )

    def sorted(self, routes: Iterable[Route], inplace: bool = False) -> List[Route]:
        return self.direction.sorted(routes, key=self.scoring, inplace=inplace)

    def nbest(self, n: int, routes: Iterable[Route], inplace: bool = False):
        return self.direction.nbest(n, routes, key=self.scoring, inplace=inplace)


class ShortestTimeRouteCriterion(RouteCriterion):

    def __init__(self, *args, **kwargs):
        super().__init__(
            direction=OptimizationDirection.MINIMIZATION,
            name='Shortest-Time',
            *args, **kwargs,
        )

    def scoring(self, route: Route) -> float:
        if not route.feasible:
            return MAX_FLOAT

        return route.duration


class LongestTimeRouteCriterion(RouteCriterion):

    def __init__(self, *args, **kwargs):
        super().__init__(
            direction=OptimizationDirection.MAXIMIZATION,
            name='Longest-Time',
            *args, **kwargs,
        )

    def scoring(self, route: Route) -> float:
        if not route.feasible:
            return MIN_FLOAT

        return route.duration


class LongestUtilTimeRouteCriterion(RouteCriterion):

    def __init__(self, *args, **kwargs):
        super().__init__(
            direction=OptimizationDirection.MAXIMIZATION,
            name='Longest-Util-Time',
            *args, **kwargs,
        )

    def scoring(self, route: Route) -> float:
        if not route.feasible:
            return MIN_FLOAT

        scoring = 0.0
        for trip in route.trips:
            scoring += trip.distance
        return scoring


class HashCodeRouteCriterion(RouteCriterion):

    def __init__(self, *args, **kwargs):
        super().__init__(
            direction=OptimizationDirection.MAXIMIZATION,
            name='Longest-Time',
            *args, **kwargs,
        )

    def scoring(self, route: Route) -> float:
        if not route.feasible:
            return MIN_FLOAT

        scoring = 0.0
        for planned_trip in route.planned_trips:
            scoring += planned_trip.distance
            if planned_trip.pickup_time == planned_trip.trip.origin_earliest:
                scoring += planned_trip.trip.on_time_bonus
        return scoring



from __future__ import annotations

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
)
from copy import deepcopy

if TYPE_CHECKING:
    from typing import (
        Set,
    )
    from ....models import (
        Planning,
        Result,
        Objective,
        Route
    )


class Breeder(ABC):

    def __init__(self, result: Result, with_copy: bool = True):
        if with_copy:
            result = deepcopy(result)

        self.result = result

    @property
    def objective(self) -> Objective:
        return self.result.objective

    @property
    def planning(self) -> Planning:
        return self.result.planning

    @property
    def routes(self) -> Set[Route]:
        return self.planning.routes

    def improve(self) -> Result:

        if __debug__:
            for route in self.routes:
                assert all(s1 == s2.previous for s1, s2 in zip(route.stops[:-1], route.stops[1:]))
                assert all(s1.departure_time <= s2.arrival_time for s1, s2 in zip(route.stops[:-1], route.stops[1:]))

        result = self._improve()

        if __debug__:
            for route in result.routes:
                assert all(s1 == s2.previous for s1, s2 in zip(route.stops[:-1], route.stops[1:]))
                if not all(s1.departure_time <= s2.arrival_time for s1, s2 in zip(route.stops[:-1], route.stops[1:])):
                    assert False

        return result

    @abstractmethod
    def _improve(self) -> Result:
        pass



import unittest
from copy import deepcopy

import jinete as jit
from tests.utils import (
    generate_one_surface,
    generate_one_position,
)


class TestGeometricPosition(unittest.TestCase):

    def setUp(self) -> None:
        self.surface = generate_one_surface()

    def test_construction(self):
        coordinates = (3, 4,)
        position = jit.GeometricPosition(coordinates, self.surface)
        self.assertEqual(coordinates[0], position[0])
        self.assertEqual(coordinates[1], position[1])
        self.assertEqual(hash(coordinates), hash(position))

    def test_equality(self):
        coordinates = (1.1, 2.2,)

        a = jit.GeometricPosition(coordinates, self.surface)
        b = jit.GeometricPosition(coordinates, self.surface)

        self.assertEqual(a, b)

    def test_not_equality(self):
        a = jit.GeometricPosition([1.1, 2.2], self.surface)
        b = jit.GeometricPosition([3.3, 4.4], self.surface)

        self.assertNotEqual(a, b)

    def test_deepcopy(self):
        position = generate_one_position()
        copied_position = deepcopy(position)
        self.assertEqual(position, copied_position)

    def test_as_tuple(self):
        coordinates = (3, 4,)
        position = jit.GeometricPosition(coordinates, self.surface)

        expected = (
            ('coordinates', coordinates),
        )
        self.assertEqual(expected, tuple(position))

    def test_as_str(self):
        coordinates = (3, 4,)
        position = jit.GeometricPosition(coordinates, self.surface)

        expected = '(' + ",".join(f"{x:07.3f}" for x in coordinates) + ')'

        self.assertEqual(expected, str(position))


if __name__ == '__main__':
    unittest.main()



from ._version import (
    __version__,
    VERSION,
)

from .models import (
    Position,
    GeometricPosition,
    Surface,
    GeometricSurface,
    DistanceMetric,
    Service,
    Trip,
    Job,
    Vehicle,
    Fleet,
    Route,
    Planning,
    PlannedTrip,

    Stop,

    Result,

    OptimizationDirection,

    Objective,
    DialARideObjective,
    TaxiSharingObjective,
    HashCodeObjective,

    RouteCriterion,
    ShortestTimeRouteCriterion,
    LongestTimeRouteCriterion,
    LongestUtilTimeRouteCriterion,
    HashCodeRouteCriterion,

    MAX_INT,
    MIN_INT,
    MAX_FLOAT,
    MIN_FLOAT,
)

from .loaders import (
    LoaderException,
    LoaderFormatterException,
    Loader,
    FileLoader,
    LoaderFormatter,
    HashCodeLoaderFormatter,
    CordeauLaporteLoaderFormatter,
)

from .dispatchers import (
    Dispatcher,
    StaticDispatcher,
)

from .storers import (
    StorerFormatter,
    ColumnarStorerFormatter,
    HashCodeStorerFormatter,
    Storer,
    FileStorer,
    PromptStorer,
    GraphPlotStorer,
    StorerSet,
)

from .algorithms import (
    Algorithm,
    NaiveAlgorithm,
    InsertionAlgorithm,
    LocalSearchAlgorithm,
    GraspAlgorithm,
    MilpAlgorithm,
    IterativeAlgorithm,
    Crosser,
    StatelessCrosser,
    BestStatelessCrosser,
    OrderedCrosser,
    RandomizedCrosser,
    Breeder,
    FlipBreeder,
    Conjecturer,
    SamplingConjecturer,
    IntensiveConjecturer,
)

from .exceptions import (
    JineteException,
    StopPlannedTripIterationException,
    NonFeasiblePlannedTripException,
    NonFeasibleRouteException,
    PreviousStopNotInRouteException,
)



import unittest

import jinete as jit

from tests.utils import (
    generate_vehicles,
)


class TestFleet(unittest.TestCase):

    def test_construction(self):
        vehicles = generate_vehicles(3)
        fleet = jit.Fleet(vehicles)

        self.assertIsInstance(fleet, jit.Fleet)
        self.assertEqual(fleet.vehicles, vehicles)

    def test_as_tuple(self):
        vehicles = generate_vehicles(3)
        fleet = jit.Fleet(vehicles)

        expected = (
            ('vehicle_identifiers', tuple(vehicle.identifier for vehicle in vehicles)),
        )
        self.assertEqual(expected, tuple(fleet))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
from time import time

from ..models import (
    Planning,
    Result,
)

if TYPE_CHECKING:
    from ..models import (
        Fleet,
        Job,
        Objective,
    )
logger = logging.getLogger(__name__)


class Algorithm(ABC):

    def __init__(self, fleet: Fleet, job: Job, *args, **kwargs):
        self.fleet = fleet
        self.job = job

    @property
    def objective(self) -> Objective:
        return self.job.objective

    def optimize(self) -> Result:
        logger.info(f'Optimizing with {self.__class__.__name__}...')

        start_time = time()
        planning = self._optimize()
        end_time = time()
        computation_time = end_time - start_time

        result = Result(
            fleet=self.fleet,
            job=self.job,
            algorithm_cls=self.__class__,
            planning=planning,
            computation_time=computation_time,
        )
        logger.info(f'Optimized with {self.__class__.__name__} obtaining {"" if result.feasible else "non "}'
                    f'feasible results!')
        return result

    @abstractmethod
    def _optimize(self) -> Planning:
        pass



from __future__ import annotations

import logging
from abc import ABC
from functools import reduce
from operator import add

from typing import (
    TYPE_CHECKING,
)
from .constants import (
    OptimizationDirection,
)

from .routes import (
    Route,
)
from .plannings import (
    Planning,
)
from .planned_trips import (
    PlannedTrip,
)
from .results import (
    Result,
)
from .stops import (
    Stop,
)

if TYPE_CHECKING:
    from typing import (
        Optional,
        TypeVar,
        Tuple,
    )

    Optimizable = TypeVar('Optimizable', Result, Planning, Route, Stop, PlannedTrip, Tuple[float, ...])

logger = logging.getLogger(__name__)


class Objective(ABC):
    direction: OptimizationDirection

    def __init__(self, name: str, dimension_count: int):
        self.name = name
        self.direction = OptimizationDirection.MAXIMIZATION
        self.dimension_count = dimension_count

    def best(self, *args: Optional[Optimizable]) -> Optimizable:
        return self.direction(
            (arg for arg in args if arg is not None),
            key=self.optimization_function,
            default=None,
        )

    def optimization_function(self, value: Optimizable) -> Tuple[float, ...]:
        if isinstance(value, Result):
            result = self._result_optimization_function(value)
        elif isinstance(value, Planning):
            result = self._planning_optimization_function(value)
        elif isinstance(value, Route):
            result = self._route_optimization_function(value)
        elif isinstance(value, Stop):
            result = self._stop_optimization_function(value)
        elif isinstance(value, PlannedTrip):
            result = self._planned_trip_optimization_function(value)
        else:
            result = value
        logger.debug(f'Computed optimization function value and obtained "{result}" from "{value}".')
        return result

    def _result_optimization_function(self, result: Result) -> Tuple[float, ...]:
        return self._planning_optimization_function(result.planning)

    def _planning_optimization_function(self, planning: Planning) -> Tuple[float, ...]:
        return reduce(
            lambda a, b: tuple(map(add, a, b)),
            (self._route_optimization_function(route) for route in planning.routes),
            tuple(0 for _ in range(self.dimension_count)),
        )

    def _route_optimization_function(self, route: Route) -> Tuple[float, ...]:
        return reduce(
            lambda a, b: tuple(map(add, a, b)),
            (self._stop_optimization_function(stop) for stop in route.stops),
            tuple(0 for _ in range(self.dimension_count)),
        )

    def _stop_optimization_function(self, stop: Stop) -> Tuple[float, ...]:
        return reduce(
            lambda a, b: tuple(map(add, a, b)),
            (self._planned_trip_optimization_function(planned_trip) for planned_trip in stop.pickups),
            tuple(0 for _ in range(self.dimension_count)),
        )

    def _planned_trip_optimization_function(self, planned_trip: PlannedTrip) -> Tuple[float, ...]:
        raise NotImplementedError


class DialARideObjective(Objective):

    def __init__(self):
        super().__init__(
            name='Dial-a-Ride',
            dimension_count=2,
        )

    def _route_optimization_function(self, route: Route) -> Tuple[float, ...]:
        scoring = 0.0
        for stop in route.stops:
            scoring -= stop.distance
        return len(tuple(route.trips)), scoring

    def _planned_trip_optimization_function(self, planned_trip: PlannedTrip) -> Tuple[float, ...]:
        scoring = 0.0
        current = planned_trip.delivery
        while current != planned_trip.pickup and current.previous is not None:
            scoring -= current.distance
            current = current.previous
        return 1, scoring


class TaxiSharingObjective(Objective):

    def __init__(self):
        super().__init__(
            name='Taxi-Sharing',
            dimension_count=1,
        )

    def _planned_trip_optimization_function(self, planned_trip: PlannedTrip) -> Tuple[float, ...]:
        if planned_trip.capacity == 0:
            return 0.0,
        return planned_trip.duration,


class HashCodeObjective(Objective):
    def __init__(self):
        super().__init__(
            name='HashCode-2018',
            dimension_count=1,
        )

    def _planned_trip_optimization_function(self, planned_trip: PlannedTrip) -> Tuple[float, ...]:
        if planned_trip.capacity == 0:
            return 0.0,
        trip = planned_trip.trip
        scoring = trip.distance
        if trip.origin_earliest == planned_trip.pickup_time:
            scoring += trip.on_time_bonus
        return scoring,



from __future__ import annotations

import unittest
from typing import TYPE_CHECKING
import jinete as jit
from .abc import (
    TestRouteCriterion,
)

if TYPE_CHECKING:
    pass


class TestLongestTimePlannedTripCriterion(TestRouteCriterion):

    def test_creation(self):
        criterion = jit.LongestTimeRouteCriterion()
        self.assertEqual(jit.OptimizationDirection.MAXIMIZATION, criterion.direction)
        self.assertEqual('Longest-Time', criterion.name)

    def test_sorting(self):
        criterion = jit.LongestTimeRouteCriterion()

        self.assertEqual(
            self.routes[::-1],
            criterion.sorted(self.routes),
        )

    def test_scoring(self):
        criterion = jit.LongestTimeRouteCriterion()

        self.assertEqual(
            4.0,
            criterion.scoring(self.routes[0]),
        )

        self.assertEqual(
            41.0,
            criterion.scoring(self.routes[1]),
        )

    def test_best(self):
        criterion = jit.LongestTimeRouteCriterion()

        self.assertEqual(
            self.routes[1],
            criterion.best(*self.routes),
        )


if __name__ == '__main__':
    unittest.main()



import unittest

import jinete as jit
from tests.utils import (
    generate_one_result,
)


class TestHashCodeStorerFormatter(unittest.TestCase):
    result: jit.Result
    expected_result: str

    @classmethod
    def setUpClass(cls) -> None:
        cls.result = generate_one_result()
        cls.expected_result = '\n'.join([
            '1 2',
            '2 0 1',
        ])

    def test_creation(self):
        storer = jit.HashCodeStorerFormatter(result=self.result)
        self.assertEqual(self.expected_result, storer.format())


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import heapq
import logging
from math import sqrt
from sys import maxsize
from typing import TYPE_CHECKING
from enum import (
    Enum,
    unique,
)

if TYPE_CHECKING:
    from typing import (
        Callable,
        Iterable,
        TypeVar,
        List,
        Union,
    )

    T = TypeVar('T')
    Number = Union[int, float]

logger = logging.getLogger(__name__)

MAX_INT = maxsize
MIN_INT = -maxsize

MAX_FLOAT = float('inf')
MIN_FLOAT = float('-inf')


@unique
class OptimizationDirection(Enum):
    MAXIMIZATION = max
    MINIMIZATION = min

    def __call__(self, *args, **kwargs):
        return self.value(*args, **kwargs)

    def __str__(self):
        return f'{self.name.capitalize()}'

    @property
    def _reverse(self):
        return self.value == max

    def sorted(self, iterable: Iterable[T], key: Callable[[T], float], inplace: bool = False) -> List[T]:
        if isinstance(iterable, list) and inplace:
            iterable.sort(key=key, reverse=self._reverse)
        else:
            if inplace:
                logger.warning('"inplace" sorting is possible only over list objects.')
            iterable = sorted(iterable, key=key, reverse=self._reverse)
        return iterable

    def nbest(self, n: int, iterable: Iterable[T], key: Callable[[T], float], inplace: bool = False) -> List[T]:
        if self._reverse:
            return heapq.nlargest(n, iterable, key=key)
        else:
            return heapq.nsmallest(n, iterable, key=key)


@unique
class DistanceMetric(Enum):

    @staticmethod
    def _euclidean_distance(a: Iterable[Number], b: Iterable[Number]) -> float:
        return sqrt(sum(pow(a_i - b_i, 2) for a_i, b_i in zip(a, b)))

    @staticmethod
    def _manhattan_distance(a: Iterable[Number], b: Iterable[Number]) -> float:
        return sum(abs(a_i - b_i) for a_i, b_i in zip(a, b))

    EUCLIDEAN = _euclidean_distance
    MANHATTAN = _manhattan_distance

    def __call__(self, *args, **kwargs):
        return self.value(*args, **kwargs)

    def __str__(self):
        return f'{self.name.capitalize()}'



from __future__ import annotations

from typing import TYPE_CHECKING

from .abc import (
    Model,
)
from .objectives import (
    Objective,
    DialARideObjective,
)

if TYPE_CHECKING:
    from typing import (
        Set,
        Any,
        Generator,
        Tuple,
        Type,
        Optional,
    )
    from .trips import (
        Trip,
    )


class Job(Model):
    trips: Set[Trip]
    objective_cls: Type[Objective]
    _objective: Optional[Objective]

    def __init__(self, trips: Set[Trip], objective_cls: Type[Objective], *args, **kwargs):
        if objective_cls is None:
            objective_cls = DialARideObjective

        self.trips = trips
        self.objective_cls = objective_cls
        self._objective = None

        self.args = args
        self.kwargs = kwargs

    @property
    def objective(self) -> Objective:
        if self._objective is None:
            self._objective = self.objective_cls(*self.args, **self.kwargs)
        return self._objective

    def __deepcopy__(self, memo) -> Job:
        return self

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        yield from (
            ('trip_identifiers', tuple(trip.identifier for trip in self.trips)),
            ('objective', self.objective),
        )



from typing import List, TypeVar, Set

E = TypeVar('E')


def remove_duplicates(seq: List[E]) -> List[E]:
    seen: Set[E] = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]



from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
)

from cached_property import cached_property

from .abc import (
    Model,
)
from .services import (
    Service,
)
from .constants import (
    MAX_FLOAT,
)

if TYPE_CHECKING:
    from typing import (
        Dict,
        Any,
        Generator,
        Tuple,
    )
    from .positions import (
        Position,
    )

logger = logging.getLogger(__name__)


class Trip(Model):
    __slots__ = (
        'identifier',
        'origin',
        'destination',
        'on_time_bonus',
        'capacity',
        'timeout',
    )
    identifier: str
    origin_position: Position
    destination_position: Position
    origin_earliest: float
    timeout: float
    on_time_bonus: float
    origin_duration: float
    capacity: float

    def __init__(self, identifier: str, origin: Service, destination: Service, capacity: float = 1,
                 on_time_bonus: float = 0.0, timeout: float = MAX_FLOAT):
        self.identifier = identifier
        self.origin = origin
        self.destination = destination
        self.on_time_bonus = on_time_bonus
        self.capacity = capacity
        self.timeout = timeout

    @property
    def origin_position(self) -> Position:
        return self.origin.position

    @property
    def origin_earliest(self) -> float:
        return self.origin.earliest

    @property
    def origin_latest(self) -> float:
        return self.origin.latest

    @property
    def origin_duration(self) -> float:
        return self.origin.duration

    @property
    def destination_position(self) -> Position:
        return self.destination.position

    @property
    def destination_earliest(self) -> float:
        return self.destination.earliest

    @property
    def destination_latest(self) -> float:
        return self.destination.latest

    @property
    def destination_duration(self) -> float:
        return self.destination.duration

    @property
    def empty(self) -> bool:
        return self.capacity == 0

    @cached_property
    def distance(self) -> float:
        return self.origin_position.distance_to(self.destination_position)

    def duration(self, now: float):
        return self.origin_position.time_to(self.destination_position, now)

    def __deepcopy__(self, memo: Dict[int, Any]) -> Trip:
        return self

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        yield from (
            ('identifier', self.identifier),
            ('origin', tuple(self.origin)),
            ('destination', tuple(self.destination)),
            ('on_time_bonus', self.on_time_bonus),
            ('capacity', self.capacity),
            ('timeout', self.timeout),
        )



from __future__ import annotations

import logging
from random import Random
from typing import TYPE_CHECKING
from .ordered import (
    OrderedCrosser,
)

if TYPE_CHECKING:
    from typing import (
        Optional,
        List,
    )
    from ....models import (
        PlannedTrip,
    )

logger = logging.getLogger(__name__)


class RandomizedCrosser(OrderedCrosser):

    def __init__(self, randomized_size: int = 3, seed: int = 56, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.randomized_size = randomized_size
        self.random = Random(seed)

    def get_planned_trip(self) -> Optional[PlannedTrip]:
        if len(self.ranking) == 0:
            return None

        candidates: List[PlannedTrip] = list()
        for sub_ranking in self.ranking.values():
            if len(sub_ranking) == 0:
                continue
            for current in sub_ranking.values():
                if len(candidates) != 0:
                    best = self.criterion.best(candidates[-1], current)
                    if not best != candidates[-1]:
                        break

                if self.randomized_size < len(candidates):
                    candidates.pop()
                candidates.append(current)
            self.criterion.sorted(candidates, inplace=True)
        if len(candidates) == 0:
            return None
        best = self.random.choice(candidates)
        return best



import unittest

import jinete as jit

from tests.utils import (
    generate_vehicles,
    generate_trips,
    generate_one_surface,
)


class TestRandomizedCrosser(unittest.TestCase):

    def test_creation(self):
        surface = generate_one_surface()
        job = jit.Job(
            trips=generate_trips(10, surface=surface),
            objective_cls=jit.DialARideObjective,
        )
        fleet = jit.Fleet(generate_vehicles(10, surface=surface))
        randomized_size = 2
        dispatcher = jit.RandomizedCrosser(
            randomized_size=randomized_size,
            job=job,
            fleet=fleet,
        )
        self.assertEqual(dispatcher.randomized_size, randomized_size)
        self.assertEqual(dispatcher.job, job)
        self.assertEqual(dispatcher.fleet, fleet)


if __name__ == '__main__':
    unittest.main()



import unittest

from tests.utils import (
    generate_one_planned_trip,
)


class TestPlannedTrip(unittest.TestCase):

    def test_feasible_planned_trip(self):
        planned_trip = generate_one_planned_trip(True)
        self.assertTrue(planned_trip.feasible)

    def test_not_feasible_planned_trip(self):
        planned_trip = generate_one_planned_trip(False)
        self.assertFalse(planned_trip.feasible)


if __name__ == '__main__':
    unittest.main()



import unittest
from copy import deepcopy

import jinete as jit

from tests.utils import (
    generate_trips,
    generate_one_job,
)


class TestJob(unittest.TestCase):

    def test_construction(self):
        trips = generate_trips(3)
        objective_cls = jit.HashCodeObjective
        job = jit.Job(trips, objective_cls)

        self.assertIsInstance(job, jit.Job)
        self.assertEqual(job.trips, trips)
        self.assertEqual(job.objective_cls, objective_cls)

    def test_deepcopy(self):
        job = generate_one_job(True)
        copied_job = deepcopy(job)
        self.assertEqual(job, copied_job)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ...models import (
    GeometricSurface,
    DistanceMetric,
    Fleet,
    Vehicle,
    Job,
    Trip,
    DialARideObjective,
    Service,
)
from .abc import (
    LoaderFormatter,
)

if TYPE_CHECKING:
    from ...models import (
        Surface,
    )

logger = logging.getLogger(__name__)


class CordeauLaporteLoaderFormatter(LoaderFormatter):

    def fleet(self, surface: Surface, *args, **kwargs) -> Fleet:
        row = self.data[0]
        m = int(row[0])

        depot_row = self.data[1]
        depot_position = surface.get_or_create_position(depot_row[1:3])

        origin = Service(depot_position)

        capacity = row[3]
        timeout = row[2]

        vehicles = set()
        for idx in range(m):
            vehicle = Vehicle(
                str(idx),
                origin,
                capacity=capacity,
                timeout=timeout,
            )

            vehicles.add(vehicle)
        fleet = Fleet(vehicles)
        logger.info(f'Created {fleet}!')
        return fleet

    def job(self, surface: Surface, *args, **kwargs) -> Job:
        row = self.data[0]
        n = int(row[1] // 2)

        trips = set()
        for idx in range(n):
            trip = self.build_trip(surface, idx, n)
            trips.add(trip)
        job = Job(trips, objective_cls=DialARideObjective)
        logger.info(f'Created "{job}"!')
        return job

    def build_trip(self, surface: Surface, idx: int, n: int) -> Trip:
        origin_idx = idx + 2
        origin_row = self.data[origin_idx]
        origin = Service(
            position=surface.get_or_create_position(origin_row[1:3]),
            earliest=origin_row[5],
            latest=origin_row[6],
            duration=origin_row[3],
        )

        destination_row = self.data[origin_idx + n]
        destination = Service(
            position=surface.get_or_create_position(destination_row[1:3]),
            earliest=destination_row[5],
            latest=destination_row[6],
            duration=destination_row[3],
        )

        identifier = f'{idx + 1:.0f}'

        assert origin_row[4] == -destination_row[4]
        capacity = origin_row[4]

        timeout = self.data[0][4]

        trip = Trip(
            identifier=identifier,
            origin=origin,
            destination=destination,
            capacity=capacity,
            timeout=timeout,
        )
        return trip

    def surface(self, *args, **kwargs) -> Surface:
        surface = GeometricSurface(DistanceMetric.EUCLIDEAN)
        logger.info(f'Created surface!')
        return surface



from __future__ import annotations

import unittest
from typing import TYPE_CHECKING
import itertools as it

import jinete as jit

from tests.utils import (
    generate_one_vehicle,
    generate_one_position,
    generate_one_planned_trip,
)

if TYPE_CHECKING:
    from typing import (
        List,
    )


class TestStop(unittest.TestCase):
    vehicle: jit.Vehicle
    position: jit.Position
    stops: List[jit.Stop]

    @classmethod
    def setUpClass(cls) -> None:
        cls.vehicle = generate_one_vehicle()
        cls.position = generate_one_position()

    def setUp(self) -> None:
        stop0 = jit.Stop(self.vehicle, self.position, None)

        stop1 = jit.Stop(self.vehicle, generate_one_position(), stop0)
        stop0.following = stop1

        stop2 = jit.Stop(self.vehicle, generate_one_position(), stop1)
        stop1.following = stop2

        stop3 = jit.Stop(self.vehicle, generate_one_position(), stop2)
        stop2.following = stop3

        stop4 = jit.Stop(self.vehicle, generate_one_position(), stop3)
        stop3.following = stop4

        stops = [stop0, stop1, stop2, stop3, stop4]
        self.stops = stops

    def test_creation(self):
        stop = jit.Stop(self.vehicle, self.position, None)

        self.assertNotIn('arrival_time', stop.__dict__)
        self.assertNotIn('departure_time', stop.__dict__)

        self.assertEqual(stop.vehicle, self.vehicle)
        self.assertEqual(stop.position, self.position)
        self.assertEqual(stop.previous, None)
        self.assertEqual(stop.previous_position, self.vehicle.origin_position)
        self.assertEqual(stop.previous_departure_time, self.vehicle.origin_earliest)
        self.assertEqual(
            stop.navigation_time,
            stop.position.time_to(self.vehicle.origin_position, stop.previous_departure_time),
        )
        self.assertEqual(stop.waiting_time, 0.0)
        self.assertEqual(stop.down_time, 0.0)
        self.assertEqual(stop.load_time, 0.0)
        self.assertEqual(0, len(tuple(stop.planned_trips)))
        self.assertEqual(0, len(tuple(stop.trips)))

    def test_merge_pickup(self):
        planned_trip = generate_one_planned_trip(True)
        other = planned_trip.pickup
        base = jit.Stop(planned_trip.vehicle, other.position, other.previous)

        self.assertNotIn(planned_trip, base.pickups)
        self.assertNotEqual(planned_trip.pickup, base)

        base.merge(other)

        self.assertIn(planned_trip, base.pickups)
        self.assertEqual(planned_trip.pickup, base)
        self.assertEqual(0, len(base.deliveries))

    def test_merge_delivery(self):
        planned_trip = generate_one_planned_trip(True)
        other = planned_trip.delivery

        base = jit.Stop(planned_trip.vehicle, other.position, other.previous)

        self.assertNotIn(planned_trip, base.deliveries)
        self.assertNotEqual(planned_trip.delivery, base)

        base.merge(other)

        self.assertIn(planned_trip, base.deliveries)
        self.assertEqual(planned_trip.delivery, base)
        self.assertEqual(0, len(base.pickups))

    def test_creation_with_previous(self):
        previous_position = generate_one_position()
        previous_stop = jit.Stop(self.vehicle, previous_position, None)

        stop = jit.Stop(self.vehicle, self.position, previous_stop)
        previous_stop.following = stop

        self.assertEqual(previous_stop.following, stop)

        self.assertEqual(stop.vehicle, self.vehicle)
        self.assertEqual(stop.position, self.position)
        self.assertEqual(stop.previous, previous_stop)
        self.assertEqual(stop.distance, stop.position.distance_to(previous_stop.position))
        self.assertEqual(stop.previous_position, previous_stop.position)
        self.assertEqual(stop.previous_departure_time, previous_stop.departure_time)
        self.assertEqual(
            stop.navigation_time,
            stop.position.time_to(previous_stop.position, stop.previous_departure_time),
        )

    def test_flush(self):
        stop = jit.Stop(self.vehicle, self.position, None)

        self.assertNotIn('arrival_time', stop.__dict__)
        self.assertNotIn('departure_time', stop.__dict__)

        self.assertIsInstance(stop.departure_time, float)

        self.assertIn('arrival_time', stop.__dict__)
        self.assertIn('departure_time', stop.__dict__)

        stop.flush()

        self.assertNotIn('arrival_time', stop.__dict__)
        self.assertNotIn('departure_time', stop.__dict__)

    def test_cache(self):
        self.assertIsInstance(self.stops[-1].departure_time, float)
        for stop in self.stops:
            self.assertIn('arrival_time', stop.__dict__)
            self.assertIn('departure_time', stop.__dict__)

    def test_all_previous(self):
        self.assertIsInstance(self.stops[-1].departure_time, float)

        self.stops[2].flush_all_previous()

        for stop in self.stops[:3]:
            self.assertNotIn('arrival_time', stop.__dict__)
            self.assertNotIn('departure_time', stop.__dict__)

        for stop in self.stops[3:]:
            self.assertIn('arrival_time', stop.__dict__)
            self.assertIn('departure_time', stop.__dict__)

    def test_with_planned_trip(self):
        stop = jit.Stop(self.vehicle, self.position, self.stops[0])

        delivery_planned_trip = generate_one_planned_trip(
            feasible=True,
            vehicle=self.vehicle,
            pickup_stop=self.stops[0],
            delivery_stop=stop,
        )

        pickup_planned_trip = generate_one_planned_trip(
            feasible=True,
            vehicle=self.vehicle,
            pickup_stop=stop,
        )
        self.assertIn(delivery_planned_trip, stop.deliveries)
        self.assertNotIn(delivery_planned_trip, stop.pickups)

        self.assertNotIn(pickup_planned_trip, stop.deliveries)
        self.assertIn(pickup_planned_trip, stop.pickups)
        self.assertIsInstance(stop.identifier, str)

        trips_seq = ''.join(
            it.chain(
                (f'P{planned_trip.trip_identifier}' for planned_trip in stop.pickups),
                (f'D{planned_trip.trip_identifier}' for planned_trip in stop.deliveries),
            )
        )
        identifier = f'{stop.position},{stop.arrival_time:.2f}:{stop.departure_time:.2f},({trips_seq})'
        self.assertEqual(identifier, stop.identifier)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import unittest
from abc import ABC
from typing import TYPE_CHECKING

import jinete as jit

if TYPE_CHECKING:
    from typing import (
        List,
    )


class TestRouteCriterion(unittest.TestCase, ABC):
    routes: List[jit.Route]

    @classmethod
    def setUpClass(cls) -> None:
        surface = jit.GeometricSurface(jit.DistanceMetric.MANHATTAN)
        origin = jit.Service(surface.get_or_create_position([0, 0]))
        vehicle = jit.Vehicle(
            identifier='TEST',
            origin=origin,
        )

        cls.routes = [
            cls._build_route_1(vehicle, surface),
            cls._build_route_2(vehicle, surface),
        ]

    @classmethod
    def _build_route_1(cls, vehicle: jit.Vehicle, surface: jit.Surface) -> jit.Route:
        route = jit.Route(vehicle)

        trip = jit.Trip(
            identifier='TEST_1',
            origin=jit.Service(
                position=surface.get_or_create_position([0, 0]),
                earliest=0.0,
                latest=10.0,
            ),
            destination=jit.Service(
                position=surface.get_or_create_position([1, 1]),
            ),
        )

        pickup_stop = jit.Stop(vehicle, surface.get_or_create_position([0, 0]), route.current_stop)
        delivery_stop = jit.Stop(vehicle, surface.get_or_create_position([1, 1]), pickup_stop)

        planned_trip = jit.PlannedTrip(
            vehicle=vehicle,
            trip=trip,
            pickup=pickup_stop,
            delivery=delivery_stop,
        )
        route.append_planned_trip(planned_trip)
        return route

    @classmethod
    def _build_route_2(cls, vehicle: jit.Vehicle, surface: jit.Surface) -> jit.Route:
        route = jit.Route(vehicle)

        trip = jit.Trip(
            identifier='TEST_2',
            origin=jit.Service(
                position=surface.get_or_create_position([0, 0]),
                earliest=1.0,
                latest=20.0,
            ),
            destination=jit.Service(
                position=surface.get_or_create_position([10, 10]),
            ),
        )

        pickup_stop = jit.Stop(vehicle, surface.get_or_create_position([0, 0]), route.current_stop)
        delivery_stop = jit.Stop(vehicle, surface.get_or_create_position([10, 10]), pickup_stop)

        planned_trip = jit.PlannedTrip(
            vehicle=vehicle,
            trip=trip,
            pickup=pickup_stop,
            delivery=delivery_stop,
        )
        route.append_planned_trip(planned_trip)
        return route



import unittest
from pathlib import Path
from unittest.mock import patch

import jinete as jit

from tests.utils import (
    generate_one_result,
)


class TestStorerSet(unittest.TestCase):
    directory_path: Path
    result: jit.Result

    @classmethod
    def setUpClass(cls) -> None:
        cls.directory_path = Path('/tmp/')
        cls.result = generate_one_result()

    def test_creation(self):
        storer_cls_set = {
            jit.PromptStorer,
            jit.GraphPlotStorer,
        }
        storer = jit.StorerSet(
            result=self.result,
            storer_cls_set=storer_cls_set,
        )
        self.assertEqual(storer.result, self.result)
        self.assertEqual(storer.storer_cls_set, storer_cls_set)

    @patch("jinete.storers.plots.graph.plt.show")
    def test_store(self, mocked_plt):
        storer = jit.StorerSet(
            result=self.result,
            storer_cls_set={
                jit.PromptStorer,
                jit.GraphPlotStorer,
            },
        )

        storer.store()


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import unittest
from operator import attrgetter
from typing import TYPE_CHECKING
import jinete as jit

if TYPE_CHECKING:
    from typing import (
        Tuple,
    )


class TestCordeauLaporteLoaderFormatter(unittest.TestCase):
    data: Tuple[Tuple[float, ...], ...]

    @classmethod
    def setUpClass(cls) -> None:
        cls.data = (
            (1.0, 4.0, 480, 6.0, 90.0),
            (0.0, -1.044, 2.000, 0.0, 0.0, 0.0, 1440.0),
            (1.0, -2.973, 6.414, 10.0, 1.0, 0.0, 1440.0),
            (2.0, -7.667, 9.934, 10.0, 1.0, 325.0, 358.0),
            (3.0, -5.476, 1.437, 10.0, -1.0, 258.0, 287.0),
            (4.0, 0.435, 1.469, 10.0, -1.0, 0.0, 1440.0),
        )

    def test_creation(self):
        formatter = jit.CordeauLaporteLoaderFormatter(self.data)
        self.assertEqual(formatter.data, self.data)

    def test_format_surface(self):
        formatter = jit.CordeauLaporteLoaderFormatter(self.data)

        surface = formatter.surface()
        self.assertIsInstance(surface, jit.GeometricSurface)
        self.assertEqual(0, len(surface.positions))

    def test_format_fleet(self):
        formatter = jit.CordeauLaporteLoaderFormatter(self.data)

        surface = formatter.surface()
        fleet = formatter.fleet(surface)
        self.assertIsInstance(fleet, jit.Fleet)
        self.assertEqual(int(self.data[0][0]), len(fleet.vehicles))

        for idx, vehicle in enumerate(fleet.vehicles):
            self.assertEqual(str(idx), vehicle.identifier)
            self.assertIsInstance(vehicle, jit.Vehicle)
            self.assertEqual(self.data[0][2], vehicle.timeout)
            self.assertEqual(self.data[0][3], vehicle.capacity)

            position = vehicle.origin_position
            self.assertEqual(vehicle.origin_position, vehicle.destination_position)
            self.assertIsInstance(position, jit.GeometricPosition)
            self.assertEqual(self.data[1][1:3], position.coordinates)

    def test_format_job(self):
        n = int(self.data[0][1] // 2)
        formatter = jit.CordeauLaporteLoaderFormatter(self.data)

        surface = formatter.surface()
        job = formatter.job(surface)
        self.assertIsInstance(job, jit.Job)
        self.assertEqual(n, len(job.trips))

        for idx, trip in enumerate(sorted(job.trips, key=attrgetter('identifier'))):
            origin_row = self.data[2 + idx]
            destination_row = self.data[2 + idx + n]

            origin_earliest, origin_latest = origin_row[-2:]
            destination_earliest, destination_latest = destination_row[-2:]

            self.assertEqual(str(idx + 1), trip.identifier)
            self.assertEqual(self.data[0][4], trip.timeout)
            self.assertIsInstance(trip, jit.Trip)
            self.assertEqual(1.0, trip.capacity)
            self.assertEqual(0.0, trip.on_time_bonus)

            self.assertEqual(origin_earliest, trip.origin_earliest)
            self.assertEqual(origin_latest, trip.origin_latest)

            self.assertEqual(destination_earliest, trip.destination_earliest)
            self.assertEqual(destination_latest, trip.destination_latest)

            self.assertIsInstance(trip.origin_position, jit.GeometricPosition)
            self.assertEqual(origin_row[1:3], trip.origin_position.coordinates)
            self.assertIsInstance(trip.destination_position, jit.GeometricPosition)
            self.assertEqual(destination_row[1:3], trip.destination_position.coordinates)


if __name__ == '__main__':
    unittest.main()



import logging
from pathlib import Path

import coloredlogs

import jinete as jit

level = logging.INFO
logging.basicConfig(level=level)
coloredlogs.install(level=level)

logger = logging.getLogger(__name__)

BASE_PATH = Path(__file__).parents[2]
DATASETS_PATH = BASE_PATH / 'res' / 'datasets'


def main():
    logger.info('Starting...')

    file_path = DATASETS_PATH / 'cordeau-laporte' / 'a2-16.txt'

    class MyLoader(jit.FileLoader):
        def __init__(self, *args, **kwargs):
            super().__init__(
                file_path=file_path,
                formatter_cls=jit.CordeauLaporteLoaderFormatter,
                *args, **kwargs,
            )

    class MyAlgorithm(jit.InsertionAlgorithm):
        def __init__(self, *args, **kwargs):
            super().__init__(
                conjecturer_cls=jit.IntensiveConjecturer,
                *args, **kwargs,
            )

    class MyStorer(jit.PromptStorer):
        def __init__(self, *args, **kwargs):
            super().__init__(
                formatter_cls=jit.ColumnarStorerFormatter,
                *args, **kwargs,
            )

    class MyStorerSet(jit.StorerSet):
        def __init__(self, *args, **kwargs):
            super().__init__(
                storer_cls_set={
                    MyStorer,
                    jit.GraphPlotStorer,
                },
                *args, **kwargs,
            )

    dispatcher = jit.StaticDispatcher(
        MyLoader,
        MyAlgorithm,
        MyStorerSet,
    )

    result = dispatcher.run()  # noqa

    logger.info('Finished...')


if __name__ == '__main__':
    main()



from __future__ import annotations
import logging
from abc import (
    ABC,
    abstractmethod,
)
from collections import (
    defaultdict,
)

from typing import (
    TYPE_CHECKING,
)

from uuid import (
    uuid4,
)

from .abc import (
    Model,
)
from .constants import (
    DistanceMetric,
)
from .positions import (
    GeometricPosition,
)

if TYPE_CHECKING:
    from typing import (
        Set,
        Any,
        Dict,
        Generator,
        Tuple,
    )
    from uuid import (
        UUID,
    )
    from .positions import (
        Position,
    )

logger = logging.getLogger(__name__)


class Surface(Model, ABC):
    uuid: UUID
    positions: Set[Position]

    def __init__(self, positions: Set[Position] = None, uuid: UUID = None, *args, **kwargs):
        if uuid is None:
            uuid = uuid4()
        if positions is None:
            positions = set()
        self.uuid = uuid
        self.positions = positions

    def get_or_create_position(self, *args, **kwargs) -> Position:
        position = self._build_position(*args, **kwargs)
        if position not in self.positions:
            self.positions.add(position)
        return position

    @abstractmethod
    def _build_position(self, *args, **kwargs):
        pass

    @abstractmethod
    def distance(self, position_a: Position, position_b: Position) -> float:
        pass

    @abstractmethod
    def time(self, position_a: Position, position_b: Position, **kwargs) -> float:
        pass

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        yield from (
            ('position_coordinates', tuple(position.coordinates for position in self.positions)),
        )


class GeometricSurface(Surface):
    cached_distance: Dict[Position, Dict[Position, float]]

    def __init__(self, metric: DistanceMetric, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.metric = metric

        self.cached_distance = defaultdict(dict)

    def _build_position(self, *args, **kwargs):
        return GeometricPosition(surface=self, *args, **kwargs)

    def distance(self, position_a: Position, position_b: Position) -> float:
        try:
            distance = self.cached_distance[position_a][position_b]
        except KeyError:
            distance = self.metric(position_a.coordinates, position_b.coordinates)
            self.cached_distance[position_a][position_b] = distance

        return distance

    def time(self, position_a: Position, position_b: Position, **kwargs) -> float:
        return self.distance(position_a, position_b)



import unittest

import jinete as jit


class TestStaticDispatcher(unittest.TestCase):

    def test_creation(self):
        dispatcher = jit.StaticDispatcher(
            jit.Loader,
            jit.Algorithm,
            jit.Storer,
        )
        self.assertEqual(dispatcher.loader_cls, jit.Loader)
        self.assertEqual(dispatcher.algorithm_cls, jit.Algorithm)
        self.assertEqual(dispatcher.storer_cls, jit.Storer)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import seaborn as sns
import networkx as nx

from ..abc import (
    Storer,
)

if TYPE_CHECKING:
    from typing import (
        Dict,
        Any,
        Tuple,
    )
    from ...models import (
        Position,
    )


class GraphPlotStorer(Storer):

    def _generate_nodes(self, edges: Dict[Tuple[Position, Position], Dict[str, Any]]) -> Dict[Position, Dict[str, Any]]:
        nodes: Dict[Position, Dict[str, Any]] = dict()
        for trip in self.trips:
            nodes[trip.origin_position] = {
                'label': f'+{trip.identifier}',
            }
            nodes[trip.destination_position] = {
                'label': f'-{trip.identifier}',
            }
        for position_pair in edges.keys():
            if position_pair[0] not in nodes:
                nodes[position_pair[0]] = {
                    'label': '',
                }
            if position_pair[1] not in nodes:
                nodes[position_pair[1]] = {
                    'label': '',
                }
        return nodes

    def _generate_edges(self) -> Dict[Tuple[Position, Position], Dict[str, Any]]:
        edges = dict()
        for route, color in zip(self.routes, sns.husl_palette(len(self.routes))):
            for first, second in zip(route.stops[:-1], route.stops[1:]):
                edges[(first.position, second.position)] = {
                    'color': color,
                    'label': '',
                }
        return edges

    def _generate_graph(self) -> nx.Graph:
        graph = nx.DiGraph()

        edges = self._generate_edges()
        graph.add_edges_from(edges.keys())
        for position_pair, metadata in edges.items():
            graph.edges[position_pair].update(metadata)

        nodes = self._generate_nodes(edges)
        graph.add_nodes_from(nodes.keys())
        for position, metadata in nodes.items():
            graph.nodes[position].update(metadata)

        return graph

    @staticmethod
    def _show_graph(graph: nx.Graph) -> None:
        import matplotlib as mpl
        mpl.rcParams['figure.dpi'] = 300

        pos = {node: node.coordinates for node in graph.nodes.keys()}

        node_labels = {node: metadata['label'] for node, metadata in graph.nodes.items()}

        edge_color = [metadata['color'] for metadata in graph.edges.values()]
        edge_labels = {edge: metadata['label'] for edge, metadata in graph.edges.items()}

        nx.draw(graph, pos=pos, edge_color=edge_color, node_size=100)
        nx.draw_networkx_labels(graph, pos, labels=node_labels, font_size=5, font_color='white')
        nx.draw_networkx_edge_labels(graph, pos, edge_labels=edge_labels)

        plt.show()

    def store(self) -> None:
        graph = self._generate_graph()
        self._show_graph(graph)



import unittest

import jinete as jit

from tests.utils import (
    generate_one_route,
    generate_trips,
)


class TestConjecturer(unittest.TestCase):

    def test_compute(self):
        conjecturer = jit.IntensiveConjecturer()
        route = generate_one_route(planned_trips_min=10)
        trips = generate_trips(5)

        conjectured_routes = conjecturer.compute(route, trips)  # noqa
        self.assertTrue(True)
        # TODO: Improve assertions


if __name__ == '__main__':
    unittest.main()



import unittest

import jinete as jit

from tests.utils import (
    generate_vehicles,
    generate_trips,
)


class TestIterativeAlgorithm(unittest.TestCase):
    job: jit.Job
    fleet: jit.Fleet

    @classmethod
    def setUpClass(cls) -> None:
        cls.job = jit.Job(generate_trips(10), objective_cls=jit.DialARideObjective)
        cls.fleet = jit.Fleet(generate_vehicles(10))

    def test_creation(self):
        algorithm = jit.IterativeAlgorithm(
            job=self.job,
            fleet=self.fleet,
        )
        self.assertEqual(algorithm.job, self.job)
        self.assertEqual(algorithm.fleet, self.fleet)

    def test_optimize(self):
        algorithm = jit.IterativeAlgorithm(
            job=self.job,
            fleet=self.fleet,
        )
        result = algorithm.optimize()

        # TODO: Properly validate  behaviour of the provided "Result" object.
        self.assertIsNotNone(result)
        self.assertIsInstance(result, jit.Result)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .models import (
        Route,
        Stop,
        PlannedTrip,
    )


class JineteException(Exception):

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class StopPlannedTripIterationException(JineteException):

    def __init__(self):
        message = f'There are no more Planned Trips to iterate over them.'
        super().__init__(message=message)


class PreviousStopNotInRouteException(JineteException):

    def __init__(self, stop: Stop):
        self.stop = stop
        message = f'Stop "{stop}" has no previous "{stop.previous}" on the belonging route "{stop.route}".'
        super().__init__(message=message)


class NonFeasiblePlannedTripException(JineteException):

    def __init__(self, planned_trip: PlannedTrip):
        assert planned_trip.feasible is False

        self.planned_trip = planned_trip
        message = f'Planned Trip "{planned_trip}" is not feasible.'
        super().__init__(message=message)


class NonFeasibleRouteException(JineteException):

    def __init__(self, route: Route):
        assert route.feasible is False

        self.route = route
        message = f'Route "{route}" is not feasible.'
        super().__init__(message=message)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import (
        Any,
        Generator,
        Tuple,
    )


class Model(ABC):
    def __repr__(self):
        return self._print(dict(self))

    def _print(self, values):
        values = ', '.join(f'{key}={value}' for key, value in values.items())
        return f'{self.__class__.__name__}({values})'

    def _print_key_value(self, key: str, value: Any):
        return f'{key}={value}'

    @abstractmethod
    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        pass



from __future__ import annotations

import logging
from copy import deepcopy
from typing import (
    TYPE_CHECKING,
)
from cached_property import cached_property
import itertools as it

from ..exceptions import (
    PreviousStopNotInRouteException,
)
from .abc import (
    Model,
)
from .planned_trips import (
    PlannedTrip,
)
from .stops import (
    Stop,
)

if TYPE_CHECKING:
    from typing import (
        List,
        Iterator,
        Any,
        Dict,
        Optional,
        Generator,
        Tuple,
    )
    from .vehicles import (
        Vehicle,
    )
    from .trips import (
        Trip,
    )
    from .positions import (
        Position,
    )

logger = logging.getLogger(__name__)


class Route(Model):
    __slots__ = (
        'vehicle',
        'stops',
    )

    vehicle: Vehicle
    stops: List[Stop]

    def __init__(self, vehicle: Vehicle, stops: List[Stop] = None):

        self.vehicle = vehicle

        if stops is None:
            first = Stop(self.vehicle, self.vehicle.origin_position, None)
            last = Stop(self.vehicle, self.vehicle.destination_position, first)
            self.stops = [
                first, last,
            ]
        else:
            self.stops = list(stops)

    def __deepcopy__(self, memo: Dict[int, Any]) -> Route:
        vehicle = deepcopy(self.vehicle, memo)

        route = Route(vehicle)
        memo[id(self)] = route

        route.stops = deepcopy(self.stops, memo)

        return route

    def clone(self, idx: int = 0) -> Route:
        i = len(self.stops)
        mapper = dict()
        mismatches = set()
        while (i > 0) and (any(mismatches) or not i < idx):
            i -= 1

            for planned_trip in self.stops[i].deliveries:
                mapper[planned_trip] = PlannedTrip(planned_trip.vehicle, planned_trip.trip)

            mismatches = (mismatches | self.stops[i].deliveries) - self.stops[i].pickups
        assert not any(mismatches)
        idx = i

        def a(s, pt: PlannedTrip):
            npt = mapper[pt]
            npt.pickup = s
            return npt

        def b(s, pt: PlannedTrip):
            npt = mapper[pt]
            npt.delivery = s
            return npt

        cloned_stops = self.stops[:idx]
        for stop in self.stops[idx:]:
            new_stop = Stop(stop.vehicle, stop.position, cloned_stops[-1] if len(cloned_stops) else None)

            pickups = {a(new_stop, pickup) for pickup in stop.pickups}
            deliveries = {b(new_stop, delivery) for delivery in stop.deliveries}

            new_stop.pickups = pickups
            new_stop.deliveries = deliveries

            cloned_stops.append(new_stop)

        cloned_route = Route(self.vehicle, cloned_stops)
        return cloned_route

    @property
    def identifier(self):
        return '|'.join(stop.identifier for stop in self.stops)

    @property
    def planned_trips(self) -> Iterator[PlannedTrip]:
        return self.deliveries

    @property
    def pickups(self) -> Iterator[PlannedTrip]:
        return it.chain.from_iterable(stop.pickups for stop in self.stops)

    @property
    def deliveries(self) -> Iterator[PlannedTrip]:
        return it.chain.from_iterable(stop.deliveries for stop in self.stops)

    @property
    def positions(self) -> Iterator[Position]:
        yield from (stop.position for stop in self.stops)

    @cached_property
    def feasible(self) -> bool:
        if not self.first_stop.position == self.vehicle.origin_position:
            return False
        if not self.vehicle.origin_earliest <= self.first_stop.arrival_time:
            return False
        if not self.last_position == self.vehicle.destination_position:
            return False
        if not self.last_departure_time <= self.vehicle.destination_latest:
            return False
        if not self.duration <= self.vehicle.timeout:
            return False
        for planned_trip in self.planned_trips:
            if not planned_trip.feasible:
                return False
        return True

    @property
    def loaded(self):
        return any(self.planned_trips)

    @property
    def trips(self) -> Iterator[Trip]:
        yield from (planned_trip.trip for planned_trip in self.planned_trips)

    @property
    def loaded_planned_trips(self) -> Iterator[PlannedTrip]:
        yield from (planned_trip for planned_trip in self.planned_trips if not planned_trip.empty)

    @property
    def loaded_trips(self) -> Iterator[Trip]:
        yield from (planned_trip.trip for planned_trip in self.loaded_planned_trips)

    @property
    def loaded_trips_count(self) -> int:
        return sum(1 for _ in self.loaded_trips)

    @property
    def first_stop(self) -> Stop:
        stop = self.stops[0]
        assert stop.previous is None
        return stop

    @property
    def first_arrival_time(self) -> float:
        return self.first_stop.arrival_time

    @property
    def first_departure_time(self) -> float:
        return self.first_stop.departure_time

    @property
    def last_stop(self) -> Stop:
        stop = self.stops[-1]
        return stop

    @property
    def last_arrival_time(self) -> float:
        return self.last_stop.arrival_time

    @property
    def last_departure_time(self) -> float:
        return self.last_stop.departure_time

    @property
    def last_position(self) -> Position:
        return self.last_stop.position

    @property
    def current_stop(self) -> Stop:
        stop = self.stops[-2]
        return stop

    @property
    def current_arrival_time(self) -> float:
        return self.current_stop.arrival_time

    @property
    def current_departure_time(self) -> float:
        return self.current_stop.departure_time

    @property
    def current_position(self) -> Position:
        return self.current_stop.position

    @property
    def duration(self) -> float:
        return self.last_departure_time - self.first_arrival_time

    @property
    def vehicle_identifier(self) -> Optional[str]:
        if self.vehicle is None:
            return None
        return self.vehicle.identifier

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        yield from (
            ('vehicle_identifier', self.vehicle_identifier),
            ('trip_identifiers', tuple(trip.identifier for trip in self.trips))
        )

    def insert_stop(self, stop: Stop) -> Stop:
        for idx in range(len(self.stops)):
            if self.stops[idx] != stop.previous:
                continue
            return self.insert_stop_at(idx + 1, stop)
        raise PreviousStopNotInRouteException(stop)

    def insert_stop_at(self, idx: int, stop: Stop) -> Stop:
        previous_stop = self.stops[idx - 1]
        following_stop = self.stops[idx] or None

        if stop == previous_stop:
            # previous_stop.merge(stop)
            return stop

        assert set(stop.pickups).isdisjoint(stop.deliveries)

        if following_stop is not None:
            following_stop.previous = stop

        self.stops.insert(idx, stop)

        for stop in self.stops[idx + 1:]:
            stop.flush()
        return stop

    def append_planned_trip(self, planned_trip: PlannedTrip):
        assert planned_trip.delivery is not None
        assert planned_trip.pickup is not None
        assert all(s1 == s2.previous for s1, s2 in zip(self.stops[:-1], self.stops[1:]))
        assert all(s1.departure_time <= s2.arrival_time for s1, s2 in zip(self.stops[:-1], self.stops[1:]))

        self.insert_stop(planned_trip.pickup)
        self.insert_stop(planned_trip.delivery)

        assert all(s1 == s2.previous for s1, s2 in zip(self.stops[:-1], self.stops[1:]))
        assert all(s1.departure_time <= s2.arrival_time for s1, s2 in zip(self.stops[:-1], self.stops[1:]))
        logger.debug(f'Append trip with "{planned_trip.trip_identifier}" identifier to route.')



import unittest

import jinete as jit

from tests.utils import (
    generate_one_position,
)


class TestVehicles(unittest.TestCase):

    def test_construction(self):
        identifier = str(0)
        service = jit.Service(
            position=generate_one_position(),
            earliest=1800,
            latest=3600,
            duration=60,
        )
        vehicle = jit.Vehicle(identifier, service)

        self.assertEqual(identifier, vehicle.identifier)
        self.assertEqual(1, vehicle.capacity)

        self.assertEqual(service, vehicle.origin)
        self.assertEqual(service.position, vehicle.origin_position)
        self.assertEqual(service.earliest, vehicle.origin_earliest)
        self.assertEqual(service.latest, vehicle.origin_latest)
        self.assertEqual(service.duration, vehicle.origin_duration)

        self.assertEqual(service, vehicle.destination)
        self.assertEqual(service.position, vehicle.destination_position)
        self.assertEqual(service.earliest, vehicle.destination_earliest)
        self.assertEqual(service.latest, vehicle.destination_latest)
        self.assertEqual(service.duration, vehicle.destination_duration)

    def test_construction_with_capacity(self):
        capacity = 3
        service = jit.Service(
            position=generate_one_position(),
            earliest=1800,
            latest=3600,
            duration=60,
        )
        identifier = str(0)
        vehicle = jit.Vehicle(identifier, service, capacity=capacity)

        self.assertEqual(identifier, vehicle.identifier)
        self.assertEqual(capacity, vehicle.capacity)

        self.assertEqual(service, vehicle.origin)
        self.assertEqual(service.position, vehicle.origin_position)
        self.assertEqual(service.earliest, vehicle.origin_earliest)
        self.assertEqual(service.latest, vehicle.origin_latest)
        self.assertEqual(service.duration, vehicle.origin_duration)

        self.assertEqual(service, vehicle.destination)
        self.assertEqual(service.position, vehicle.destination_position)
        self.assertEqual(service.earliest, vehicle.destination_earliest)
        self.assertEqual(service.latest, vehicle.destination_latest)
        self.assertEqual(service.duration, vehicle.destination_duration)

    def test_vehicle_with_final(self):
        identifier = str(0)
        origin = jit.Service(
            position=generate_one_position(),
            earliest=100,
            latest=200,
            duration=2,
        )
        destination = jit.Service(
            position=generate_one_position(),
            earliest=1000,
            latest=2000,
            duration=20,
        )
        vehicle = jit.Vehicle(identifier, origin, destination)

        self.assertEqual(identifier, vehicle.identifier)
        self.assertEqual(1, vehicle.capacity)

        self.assertEqual(origin, vehicle.origin)
        self.assertEqual(origin.position, vehicle.origin_position)
        self.assertEqual(origin.earliest, vehicle.origin_earliest)
        self.assertEqual(origin.latest, vehicle.origin_latest)
        self.assertEqual(origin.duration, vehicle.origin_duration)

        self.assertEqual(destination, vehicle.destination)
        self.assertEqual(destination.position, vehicle.destination_position)
        self.assertEqual(destination.earliest, vehicle.destination_earliest)
        self.assertEqual(destination.latest, vehicle.destination_latest)
        self.assertEqual(destination.duration, vehicle.destination_duration)

    def test_as_tuple(self):
        identifier = str(0)
        origin = jit.Service(
            position=generate_one_position(),
            earliest=100,
            latest=200,
            duration=2,
        )
        destination = jit.Service(
            position=generate_one_position(),
            earliest=1000,
            latest=2000,
            duration=20,
        )
        capacity = 44
        vehicle = jit.Vehicle(identifier, origin, destination, capacity)

        expected = (
            ('identifier', identifier),
            ('origin', tuple(origin)),
            ('destination', tuple(destination)),
            ('capacity', capacity),
            ('timeout', vehicle.timeout),
        )

        self.assertEqual(expected, tuple(vehicle))


if __name__ == '__main__':
    unittest.main()



import unittest

import jinete as jit

from tests.utils import (
    generate_one_position,
)


class TestTrips(unittest.TestCase):

    def test_trip(self):
        identifier = str()
        origin = jit.Service(
            position=generate_one_position(),
            earliest=0.0,
        )
        destination = jit.Service(
            position=generate_one_position()
        )

        distance = origin.distance_to(destination)
        duration = origin.time_to(destination)

        trip = jit.Trip(identifier=identifier, origin=origin, destination=destination)

        self.assertEqual(origin.position, trip.origin_position)
        self.assertEqual(destination.position, trip.destination_position)
        self.assertEqual(0.0, trip.origin_earliest)
        self.assertEqual(trip.origin_latest, jit.MAX_FLOAT)
        self.assertEqual(0, trip.origin_duration)
        self.assertEqual(1, trip.capacity)
        self.assertEqual(distance, trip.distance)
        self.assertEqual(duration, trip.duration(trip.origin_earliest))

    def test_trip_with_capacity(self):
        identifier = str()
        capacity = 3
        origin = jit.Service(
            position=generate_one_position(),
            earliest=0.0,
        )
        destination = jit.Service(
            position=generate_one_position()
        )

        distance = origin.distance_to(destination)
        duration = origin.time_to(destination)

        trip = jit.Trip(identifier=identifier, origin=origin, destination=destination, capacity=capacity)

        self.assertEqual(origin.position, trip.origin_position)
        self.assertEqual(destination.position, trip.destination_position)
        self.assertEqual(0, trip.origin_earliest)
        self.assertEqual(trip.origin_latest, jit.MAX_FLOAT)
        self.assertEqual(0, trip.origin_duration)
        self.assertEqual(capacity, trip.capacity)
        self.assertEqual(distance, trip.distance)
        self.assertEqual(duration, trip.duration(trip.origin_earliest))

    def test_trip_with_timeout(self):
        identifier = str()
        origin = jit.Service(
            position=generate_one_position(),
            earliest=0.0,
            latest=3600,
        )
        destination = jit.Service(
            position=generate_one_position()
        )
        distance = origin.distance_to(destination)
        duration = origin.time_to(destination)

        trip = jit.Trip(identifier=identifier, origin=origin, destination=destination)

        self.assertEqual(origin.position, trip.origin_position)
        self.assertEqual(destination.position, trip.destination_position)
        self.assertEqual(0, trip.origin_earliest)
        self.assertEqual(3600, trip.origin_latest)
        self.assertEqual(0, trip.origin_duration)
        self.assertEqual(1, trip.capacity)
        self.assertEqual(distance, trip.distance)
        self.assertEqual(duration, trip.duration(trip.origin_earliest))

    def test_trip_with_load_time(self):
        identifier = str()
        origin = jit.Service(
            position=generate_one_position(),
            earliest=0.0,
            duration=1800,
        )
        destination = jit.Service(
            position=generate_one_position()
        )
        distance = origin.distance_to(destination)
        duration = origin.time_to(destination)

        trip = jit.Trip(identifier=identifier, origin=origin, destination=destination)

        self.assertEqual(origin.position, trip.origin_position)
        self.assertEqual(destination.position, trip.destination_position)
        self.assertEqual(trip.origin_latest, jit.MAX_FLOAT)
        self.assertEqual(1800, trip.origin_duration)
        self.assertEqual(1, trip.capacity)
        self.assertEqual(0.0, trip.origin_earliest)
        self.assertEqual(distance, trip.distance)
        self.assertEqual(duration, trip.duration(trip.origin_earliest))

    def test_as_tuple(self):
        identifier = str(0)
        origin = jit.Service(
            position=generate_one_position(),
            earliest=100,
            latest=200,
            duration=2,
        )
        destination = jit.Service(
            position=generate_one_position(),
            earliest=1000,
            latest=2000,
            duration=20,
        )
        capacity = 44
        on_time_bonus = 3
        timeout = 100
        trip = jit.Trip(
            identifier=identifier,
            origin=origin,
            destination=destination,
            capacity=capacity,
            on_time_bonus=on_time_bonus,
            timeout=timeout,
        )

        expected = (
            ('identifier', identifier),
            ('origin', tuple(origin)),
            ('destination', tuple(destination)),
            ('on_time_bonus', on_time_bonus),
            ('capacity', capacity),
            ('timeout', trip.timeout),
        )

        self.assertEqual(expected, tuple(trip))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import unittest
from abc import ABC
from typing import TYPE_CHECKING

import jinete as jit

if TYPE_CHECKING:
    pass


class TestObjective(unittest.TestCase, ABC):
    planned_trip: jit.PlannedTrip
    stop: jit.Stop
    route: jit.Route
    planning: jit.Planning
    result: jit.Result

    @classmethod
    def setUpClass(cls) -> None:
        surface = jit.GeometricSurface(jit.DistanceMetric.MANHATTAN)
        origin = jit.Service(surface.get_or_create_position([0, 0]))
        vehicle = jit.Vehicle(
            identifier='TEST',
            origin=origin,
        )
        fleet = jit.Fleet({vehicle})

        trips = [
            jit.Trip(
                identifier='TEST_1',
                origin=jit.Service(
                    position=surface.get_or_create_position([0, 0]),
                    earliest=0.0,
                    latest=10.0,
                ),
                destination=jit.Service(
                    position=surface.get_or_create_position([1, 1]),
                ),
            ),
            jit.Trip(
                identifier='TEST_1',
                origin=jit.Service(
                    position=surface.get_or_create_position([1, 1]),
                    earliest=0.0,
                    latest=20.0,
                ),
                destination=jit.Service(
                    position=surface.get_or_create_position([10, 10]),
                ),
            ),
        ]
        job = jit.Job(set(trips), jit.DialARideObjective)
        route = jit.Route(vehicle)

        pickup_stop = jit.Stop(vehicle, surface.get_or_create_position([0, 0]), route.current_stop)
        delivery_stop = jit.Stop(vehicle, surface.get_or_create_position([1, 1]), pickup_stop)

        cls.planned_trip = jit.PlannedTrip(
            vehicle=vehicle,
            trip=trips[0],
            pickup=pickup_stop,
            delivery=delivery_stop,
        )
        route.append_planned_trip(cls.planned_trip)

        cls.stop = route.stops[1]
        conjecture = jit.Conjecturer()
        cls.route = conjecture.compute_one(route, trips[1])
        cls.planning = jit.Planning({cls.route})
        cls.result = jit.Result(fleet, job, jit.NaiveAlgorithm, cls.planning, 0.0)



from __future__ import annotations

import logging
from copy import deepcopy
from functools import reduce
from operator import and_
from typing import TYPE_CHECKING
from uuid import (
    uuid4,
)
from .abc import (
    Model,
)

if TYPE_CHECKING:
    from typing import (
        Set,
        Dict,
        Any,
        Iterator,
        Generator,
        Tuple,
    )
    from .routes import (
        Route,
    )
    from .trips import (
        Trip,
    )
    from .planned_trips import (
        PlannedTrip,
    )
    from .vehicles import (
        Vehicle,
    )
    from uuid import (
        UUID,
    )

logger = logging.getLogger(__name__)


class Planning(Model):
    routes: Set[Route]
    uuid: UUID

    def __init__(self, routes: Set[Route] = None, uuid: UUID = None):
        if uuid is None:
            uuid = uuid4()
        if routes is None:
            routes = set()

        self.routes = routes
        self.uuid = uuid

    @property
    def loaded_routes(self):
        return set(route for route in self.routes if route.loaded)

    @property
    def vehicles(self) -> Iterator[Vehicle]:
        for route in self.routes:
            yield route.vehicle

    @property
    def planned_trips(self) -> Iterator[PlannedTrip]:
        for route in self.routes:
            yield from route.planned_trips

    @property
    def trips(self) -> Iterator[Trip]:
        yield from (planned_trip.trip for planned_trip in self.planned_trips)

    @property
    def feasible(self) -> bool:
        return reduce(and_, (route.feasible for route in self.routes))

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        yield from (
            ('uuid', self.uuid),
            ('route_identifiers', tuple(route.identifier for route in self.routes))
        )

    def __deepcopy__(self, memo: Dict[int, Any]) -> Planning:
        planning = Planning()
        memo[id(self)] = planning

        planning.routes = deepcopy(self.routes, memo)

        return planning



import logging

from ...models import (
    Fleet,
    Job,
    Surface,
    GeometricSurface,
    DistanceMetric,
    Vehicle,
    Trip,
    HashCodeObjective,
    Service,
)
from .abc import (
    LoaderFormatter,
)

logger = logging.getLogger(__name__)


class HashCodeLoaderFormatter(LoaderFormatter):

    def fleet(self, surface: Surface, *args, **kwargs) -> Fleet:
        row = self.data[0]
        n, latest, capacity = int(row[2]), row[5], 1.0

        origin = Service(surface.get_or_create_position([0, 0]), latest=latest)
        vehicles = set(Vehicle(str(idx), origin, capacity=capacity) for idx in range(n))
        fleet = Fleet(vehicles)
        logger.info(f'Created {fleet}!')
        return fleet

    def job(self, surface: Surface, *args, **kwargs) -> Job:
        bonus = self.data[0][4]
        rows = self.data[1:]
        trips = set(self._build_trip(surface, str(i), bonus, *row) for i, row in enumerate(rows))

        kwargs['objective_cls'] = HashCodeObjective
        job = Job(trips, *args, **kwargs)

        logger.info(f'Created "{job}"!')
        return job

    def _build_trip(self, surface: Surface, identifier: str, bonus: float, x1: float, y1: float, x2: float, y2: float,
                    earliest: float, latest: float) -> Trip:
        origin = Service(
            position=surface.get_or_create_position([x1, y1]),
            earliest=earliest,
            latest=latest,
        )
        destination = Service(
            position=surface.get_or_create_position([x2, y2]),
        )
        trip = Trip(identifier, on_time_bonus=bonus, origin=origin, destination=destination)
        logger.debug(f'Created trip!')
        return trip

    def surface(self, *args, **kwargs) -> Surface:
        # row = self.data[0]
        # rows = row[0]
        # columns = row[1]
        surface = GeometricSurface(DistanceMetric.MANHATTAN)
        logger.info(f'Created surface!')
        return surface



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from .constants import (
    MAX_FLOAT,
)
from .abc import (
    Model,
)
from .services import (
    Service,
)

if TYPE_CHECKING:
    from typing import (
        Set,
        Any,
        Dict,
        Generator,
        Tuple,
    )
    from .positions import (
        Position,
    )

logger = logging.getLogger(__name__)


class Vehicle(Model):
    identifier: str
    origin: Service
    destination: Service
    capacity: float

    def __init__(self, identifier: str, origin: Service, destination: Service = None, capacity: float = 1.0,
                 timeout: float = MAX_FLOAT):
        self.identifier = identifier

        self.origin = origin
        self._destination = destination
        self.capacity = capacity
        self.timeout = timeout

    @property
    def origin_position(self) -> Position:
        return self.origin.position

    @property
    def origin_earliest(self) -> float:
        return self.origin.earliest

    @property
    def origin_latest(self) -> float:
        return self.origin.latest

    @property
    def origin_duration(self) -> float:
        return self.origin.duration

    @property
    def destination(self) -> Service:
        if self._destination is None:
            return self.origin
        return self._destination

    @property
    def destination_position(self) -> Position:
        return self.destination.position

    @property
    def destination_earliest(self) -> float:
        return self.destination.earliest

    @property
    def destination_latest(self) -> float:
        return self.destination.latest

    @property
    def destination_duration(self) -> float:
        return self.destination.duration

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        yield from (
            ('identifier', self.identifier),
            ('origin', tuple(self.origin)),
            ('destination', tuple(self.destination)),
            ('capacity', self.capacity),
            ('timeout', self.timeout),
        )

    def __deepcopy__(self, memo: Dict[int, Any]) -> Vehicle:
        return self


class Fleet(Model):
    vehicles: Set[Vehicle]

    def __init__(self, vehicles: Set[Vehicle]):
        self.vehicles = vehicles

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        yield from (
            ('vehicle_identifiers', tuple(vehicle.identifier for vehicle in self.vehicles)),
        )

    def __deepcopy__(self, memo: Dict[int, Any]) -> Fleet:
        return self



import unittest

import jinete as jit

from tests.utils import (
    generate_one_route,
    generate_trips,
    generate_one_trip,
)


class TestConjecturer(unittest.TestCase):

    def test_compute_one(self):
        conjecturer = jit.Conjecturer()
        route = generate_one_route()
        trip = generate_one_trip()

        conjectured_route = conjecturer.compute_one(route, trip)

        self.assertIsInstance(conjectured_route, jit.Route)
        # TODO: Improve assertions

    def test_compute(self):
        conjecturer = jit.Conjecturer()
        route = generate_one_route()
        trips = generate_trips(5)

        conjectured_routes = conjecturer.compute(route, trips)
        self.assertEqual(len(trips), len(conjectured_routes))
        # TODO: Improve assertions


if __name__ == '__main__':
    unittest.main()



import unittest
from pathlib import Path

import jinete as jit
from tests.utils import (
    generate_one_result,
)


class TestFileStorer(unittest.TestCase):
    directory_path: Path
    file_path: Path
    result: jit.Result

    @classmethod
    def setUpClass(cls) -> None:
        cls.directory_path = Path('/tmp/')
        cls.file_path = cls.directory_path / 'jinete_store_result_test.txt'
        cls.result = generate_one_result()

    @classmethod
    def tearDown(cls) -> None:
        if cls.file_path.exists():
            cls.file_path.unlink()

    def test_creation(self):
        storer = jit.FileStorer(
            file_path=self.file_path,
            result=self.result,
        )
        self.assertEqual(storer.file_path, self.file_path)
        self.assertEqual(storer.result, self.result)
        self.assertEqual(storer.formatter_cls, jit.ColumnarStorerFormatter)

    def test_store(self):
        storer = jit.FileStorer(
            file_path=self.file_path,
            result=self.result,
            formatter_cls=jit.HashCodeStorerFormatter,
        )
        storer.store()

        expected_result = '\n'.join([
            "1 2",
            "2 0 1",
        ])

        self.assertEqual(self.file_path.exists(), True)
        with self.file_path.open() as file:
            result = file.read()
            self.assertEqual(expected_result, result)


if __name__ == '__main__':
    unittest.main()



import unittest
from sys import maxsize

import jinete as jit


class TestConstants(unittest.TestCase):

    def test_max_int(self):
        self.assertEqual(jit.MAX_INT, maxsize)

    def test_min_int(self):
        self.assertEqual(jit.MIN_INT, -maxsize)

    def test_max_float(self):
        self.assertEqual(jit.MAX_FLOAT, float('inf'))

    def test_min_float(self):
        self.assertEqual(jit.MIN_FLOAT, float('-inf'))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from .abc import (
    Storer,
)

if TYPE_CHECKING:
    from typing import (
        Set,
        Type,
    )

logger = logging.getLogger(__name__)


class StorerSet(Storer):

    def __init__(self, storer_cls_set: Set[Type[Storer]], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.storer_cls_set = storer_cls_set
        self.args = args
        self.kwargs = kwargs

    def store(self) -> None:
        for storer_cls in self.storer_cls_set:
            logger.info(f'Storing result with "{storer_cls.__name__}"...')

            storer = storer_cls(*self.args, **self.kwargs)
            storer.store()



from __future__ import annotations

import logging
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
)

from .abc import (
    Model,
)

if TYPE_CHECKING:
    from typing import (
        Tuple,
        Sequence,
        Generator,
        Dict,
        Any,
    )
    from .surfaces import (
        Surface,
    )

logger = logging.getLogger(__name__)


class Position(Model, ABC):
    def __init__(self, surface: Surface):
        self.surface = surface

    def distance_to(self, other: Position) -> float:
        return self.surface.distance(self, other)

    @property
    @abstractmethod
    def coordinates(self) -> Tuple[Any]:
        pass

    def time_to(self, other: Position, now: float = None) -> float:
        return self.surface.time(self, other, now=now)

    def __deepcopy__(self, memo: Dict[int, Any]) -> Position:
        return self


class GeometricPosition(Position):
    __slots__ = (
        'coordinates',
    )

    coordinates: Tuple[float, ...]

    def __init__(self, coordinates: Sequence[float], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.coordinates = tuple(coordinates)

    def __hash__(self):
        return hash(self.coordinates)

    def __eq__(self, other) -> bool:
        return self.coordinates == other.coordinates

    def __ne__(self, other) -> bool:
        return self.coordinates != other.coordinates

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        yield from (
            ('coordinates', self.coordinates),
        )

    def __str__(self):
        c = ",".join(f"{x:07.3f}" for x in self.coordinates)
        return f'({c})'

    def __getitem__(self, item):
        return self.coordinates[item]



from __future__ import annotations

import unittest
from typing import (
    TYPE_CHECKING,
)

import jinete as jit

from tests.utils import (
    generate_one_planning,
)

if TYPE_CHECKING:
    from typing import (
        Type,
    )


class TestResult(unittest.TestCase):
    planning: jit.Planning
    fleet: jit.Fleet
    job: jit.Job
    algorithm_cls: Type[jit.Algorithm]
    computation_time: float

    @classmethod
    def setUpClass(cls) -> None:
        cls.planning = generate_one_planning()
        cls.fleet = jit.Fleet(set(cls.planning.vehicles))
        cls.job = jit.Job(set(cls.planning.trips), jit.DialARideObjective)
        cls.algorithm_cls = jit.Algorithm
        cls.computation_time = 1

    def test_construction(self):
        result = jit.Result(
            fleet=self.fleet,
            job=self.job,
            algorithm_cls=self.algorithm_cls,
            planning=self.planning,
            computation_time=self.computation_time
        )

        self.assertIsInstance(result, jit.Result)
        self.assertEqual(self.fleet, result.fleet)
        self.assertEqual(self.fleet.vehicles, result.vehicles)

        self.assertEqual(self.job, result.job)
        self.assertEqual(self.job.trips, result.trips)
        self.assertEqual(self.job.objective, result.objective)
        self.assertEqual(self.job.objective.optimization_function(result), result.optimization_function)
        self.assertEqual(self.job.objective.direction, result.direction)

        self.assertEqual(self.algorithm_cls, result.algorithm_cls)

        self.assertEqual(self.planning, result.planning)
        self.assertEqual(self.planning.uuid, result.planning_uuid)
        self.assertEqual(self.planning.routes, result.routes)

        self.assertEqual(self.computation_time, result.computation_time)

    def test_as_tuple(self):
        result = jit.Result(
            fleet=self.fleet,
            job=self.job,
            algorithm_cls=self.algorithm_cls,
            planning=self.planning,
            computation_time=self.computation_time
        )

        expected = (
            ('fleet_uuid', tuple(self.fleet)),
            ('job', tuple(self.job)),
            ('algorithm_name', self.algorithm_cls.__name__),
            ('planning_uuid', self.planning.uuid)
        )
        self.assertEqual(expected, tuple(result))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ...models import (
    Planning,
)
from ...exceptions import (
    StopPlannedTripIterationException,
)
from ..abc import (
    Algorithm,
)
from ..utils import (
    OrderedCrosser,
)

if TYPE_CHECKING:
    from typing import (
        Type,
    )
    from ..utils import (
        Crosser,
    )

logger = logging.getLogger(__name__)


class InsertionAlgorithm(Algorithm):

    def __init__(self, crosser_cls: Type[Crosser] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if crosser_cls is None:
            crosser_cls = OrderedCrosser
        self.crosser_cls = crosser_cls
        self.args = args
        self.kwargs = kwargs

    def build_crosser(self) -> Crosser:
        return self.crosser_cls(*self.args, **self.kwargs)

    def _optimize(self) -> Planning:
        crosser = self.build_crosser()

        while not crosser.completed:
            try:
                route = next(crosser)
            except StopPlannedTripIterationException:
                break
            crosser.set_route(route)

        planning = Planning(crosser.routes)
        return planning



from __future__ import annotations

import logging
from typing import TYPE_CHECKING
import itertools as it

from cached_property import (
    cached_property,
)

from .abc import (
    Crosser,
)

if TYPE_CHECKING:
    from typing import (
        Optional,
        Iterator,
    )
    from ....models import (
        PlannedTrip,
        Trip,
    )

logger = logging.getLogger(__name__)


class StatelessCrosser(Crosser):

    def flush(self):
        for key in ('iterator',):
            self.__dict__.pop(key, None)

    @cached_property
    def iterator(self) -> Iterator[PlannedTrip]:
        for route, trip in it.product(self.attractive_routes, self.pending_trips):
            logger.debug(f'Yielding ({route}, {trip})...')
            planned_trip = self.conjecturer.compute_one(route, trip)
            yield planned_trip

    def get_planned_trip(self) -> Optional[PlannedTrip]:
        return next(self.iterator, None)

    def mark_trip_as_done(self, trip: Trip):
        super().mark_trip_as_done(trip)
        self.flush()



from __future__ import annotations

import unittest

from typing import TYPE_CHECKING

import jinete as jit

if TYPE_CHECKING:
    from typing import (
        Tuple,
    )


class TestHashCodeLoaderFormatter(unittest.TestCase):
    data: Tuple[Tuple[float, ...], ...]

    @classmethod
    def setUpClass(cls) -> None:
        cls.data = (
            (3.0, 4.0, 2.0, 3.0, 2.0, 10.0),
            (0.0, 0.0, 1.0, 3.0, 2.0, 9.0),
            (1.0, 2.0, 1.0, 0.0, 0.0, 9.0),
            (2.0, 0.0, 2.0, 2.0, 0.0, 9.0),
        )

    def test_creation(self):
        formatter = jit.HashCodeLoaderFormatter(self.data)
        self.assertEqual(formatter.data, self.data)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
)

import pulp as lp

if TYPE_CHECKING:
    from typing import (
        List,
        Set,
    )
    from ....models import (
        Fleet,
        Job,
        Route,
    )


class Model(ABC):

    def __init__(self, fleet: Fleet, job: Job, *args, **kwargs):
        self.fleet = fleet
        self.job = job

    @property
    @abstractmethod
    def problem(self) -> lp.LpProblem:
        pass

    @property
    @abstractmethod
    def objective(self) -> lp.LpConstraintVar:
        pass

    @property
    @abstractmethod
    def constraints(self) -> List[lp.LpConstraint]:
        pass

    @abstractmethod
    def solve(self) -> Set[Route]:
        pass



from .exceptions import (
    LoaderException,
)
from .formatters import (
    LoaderFormatterException,
    LoaderFormatter,
    HashCodeLoaderFormatter,
    CordeauLaporteLoaderFormatter,
)
from .abc import (
    Loader,
)
from .file import (
    FileLoader,
)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from abc import ABC, abstractmethod

if TYPE_CHECKING:
    from typing import (
        Type,
        Set,
    )
    from ..models import (
        Result,
        Route,
        Trip,
    )
    from .formatters import (
        StorerFormatter,
    )

logger = logging.getLogger(__name__)


class Storer(ABC):

    def __init__(self, result: Result, formatter_cls: Type[StorerFormatter] = None):
        if formatter_cls is None:
            from .formatters import ColumnarStorerFormatter
            formatter_cls = ColumnarStorerFormatter
        self.result = result
        self.formatter_cls = formatter_cls

    def formatted_planning(self) -> str:
        return self.formatter_cls(result=self.result).format()

    @property
    def trips(self) -> Set[Trip]:
        return self.result.trips

    @property
    def routes(self) -> Set[Route]:
        return self.result.routes

    @abstractmethod
    def store(self) -> None:
        pass



from __future__ import annotations

from typing import TYPE_CHECKING

from ..models import (
    Planning,
)
from .abc import (
    Algorithm,
)

if TYPE_CHECKING:
    from typing import (
        Set,
    )
    from ..models import (
        Route
    )


class NaiveAlgorithm(Algorithm):

    def _optimize(self) -> Planning:
        routes: Set[Route] = set()
        return Planning(routes)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ...models import (
    Planning,
)

from ..abc import (
    Algorithm,
)
from ..utils import (
    ThreeIndexModel,
)

if TYPE_CHECKING:
    from typing import (
        Type
    )
    from ..utils import (
        Model,
    )

logger = logging.getLogger(__name__)


class MilpAlgorithm(Algorithm):

    def __init__(self, model_cls: Type[Model] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if model_cls is None:
            model_cls = ThreeIndexModel

        self.model_cls = model_cls

        self.args = args
        self.kwargs = kwargs

    def build_model(self) -> Model:
        return self.model_cls(*self.args, **self.kwargs)

    def _optimize(self) -> Planning:
        model = self.build_model()

        routes = model.solve()

        return Planning(routes)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ...models import (
    Planning,
)
from ..abc import (
    Algorithm,
)
from ..utils import (
    FlipBreeder
)

if TYPE_CHECKING:
    from typing import (
        Set,
    )
    from ...models import (
        Result,
        Route,
    )

logger = logging.getLogger(__name__)


class LocalSearchAlgorithm(Algorithm):

    def __init__(self, initial: Result, no_improvement_threshold: int = 10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.initial = initial
        self.args = args
        self.kwargs = kwargs
        self.breeder_cls = FlipBreeder
        self.no_improvement_threshold = no_improvement_threshold

    @property
    def initial_planning(self) -> Planning:
        return self.initial.planning

    @property
    def initial_routes(self) -> Set[Route]:
        return self.initial_planning.routes

    def _optimize(self) -> Planning:
        best = self.initial

        no_improvement_count = 0
        while no_improvement_count < self.no_improvement_threshold:
            no_improvement_count += 1

            current = self.breeder_cls(best).improve()
            best = self.objective.best(best, current)

            if best == current:
                no_improvement_count = 0

        assert best is not None

        return best.planning



import unittest
from pathlib import Path
from unittest.mock import patch

import networkx as nx

import jinete as jit

from tests.utils import (
    generate_one_result,
)


class TestGraphPlotStorer(unittest.TestCase):
    directory_path: Path
    result: jit.Result

    @classmethod
    def setUpClass(cls) -> None:
        cls.directory_path = Path('/tmp/')
        cls.result = generate_one_result()

    def test_creation(self):
        storer = jit.GraphPlotStorer(
            result=self.result,
        )
        self.assertEqual(storer.result, self.result)

    @patch("jinete.storers.plots.graph.plt.show")
    def test_store(self, mocked_plt):
        storer = jit.GraphPlotStorer(
            result=self.result,
        )
        graph = storer._generate_graph()
        self.assertIsInstance(graph, nx.DiGraph)

        storer.store()


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import unittest
from typing import TYPE_CHECKING
from pathlib import Path

import jinete as jit

if TYPE_CHECKING:
    from typing import (
        Sequence
    )


class TestFileLoader(unittest.TestCase):
    directory_path: Path
    file_path: Path
    data: Sequence[Sequence[float]]

    @classmethod
    def setUpClass(cls) -> None:
        cls.directory_path = Path('/tmp/')
        cls.file_path = cls.directory_path / 'jinete_problem_test.txt'
        cls.data = [
            [1.0, 1, 480, 6.0, 90.0],
            [0.0, -1.044, 2.000, 0.0, 0.0, 0.0, 1440.0],
            [1.0, -2.973, 6.414, 10.0, 1.0, 0.0, 1440.0],
            [2.0, -5.476, 1.437, 10.0, -1.0, 258.0, 287.0],
        ]
        with cls.file_path.open('w') as file:
            file.writelines('\t'.join(str(cell) for cell in row) + '\n' for row in cls.data)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.file_path.unlink()

    def test_creation(self):
        loader = jit.FileLoader(
            file_path=self.file_path,
        )
        self.assertEqual(loader.file_path, self.file_path)
        self.assertEqual(loader.data, self.data)
        self.assertIsInstance(loader.fleet, jit.Fleet)
        self.assertIsInstance(loader.job, jit.Job)
        self.assertIsInstance(loader.surface, jit.GeometricSurface)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from distutils.util import convert_path
import itertools as it

module_name = 'jinete'

with open('README.md') as f:
    long_description = f.read()

with open(convert_path('{}/_version.py'.format(module_name))) as file:
    main_ns = dict()
    exec(file.read(), main_ns)
    module_version = main_ns['__version__']

description = (
    'High Performance solving suite for the Pickup and Delivery Problem and its related extensions.'
)

dependencies = [
    'cached-property>=1.5.1',
    'networkx>=2.3',
    'matplotlib>=3.1.1',
    'seaborn>=0.9.0',
    'PuLP>=1.6.10',
]

extra_dependencies = {
    'docs': [
        'sphinx',
        'sphinx-rtd-theme',
        'sphinxcontrib-apidoc',
    ],
    'tests': [
        'coverage',
        'codecov',
    ],
    'syntax': [
        'flake8',
        'mypy',
    ],
    'logs': [
        'coloredlogs',
    ]
}
extra_dependencies['all'] = list(it.chain(extra_dependencies.values()))

setup(
    name=module_name,
    version=module_version,
    url='https://github.com/garciparedes/jinete',
    author='Sergio García Prado',
    author_email='sergio@garciparedes.me',
    long_description=long_description,
    long_description_content_type='text/markdown',
    description=description,
    packages=find_packages(),
    install_requires=dependencies,
    extras_require=extra_dependencies,
)



import unittest

import jinete as jit
from .abc import (
    TestObjective,
)


class TestDialARideObjective(TestObjective):

    def test_creation(self):
        objective = jit.DialARideObjective()
        self.assertEqual('Dial-a-Ride', objective.name)

    def test_result(self):
        objective = jit.DialARideObjective()

        self.assertEqual(
            (2, -40.0),
            objective.optimization_function(self.result),
        )

    def test_planning(self):
        objective = jit.DialARideObjective()

        self.assertEqual(
            (2, -40.0),
            objective.optimization_function(self.planning),
        )

    def test_route(self):
        objective = jit.DialARideObjective()

        self.assertEqual(
            (2, -40.0),
            objective.optimization_function(self.route),
        )

    def test_stop(self):
        objective = jit.DialARideObjective()

        self.assertEqual(
            (1, -2.0),
            objective.optimization_function(self.stop),
        )

    def test_planning_trip(self):
        objective = jit.DialARideObjective()

        self.assertEqual(
            (1, -2.0),
            objective.optimization_function(self.planned_trip),
        )

    def test_best(self):
        objective = jit.DialARideObjective()

        self.assertEqual(
            next(self.route.planned_trips),
            objective.best(*self.route.planned_trips),
        )


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
)

from cached_property import (
    cached_property,
)

from .abc import (
    Model,
)
from .trips import (
    Trip,
)
from .stops import (
    Stop,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        Generator,
        Tuple,
    )
    from .positions import (
        Position,
    )
    from .vehicles import (
        Vehicle,
    )

logger = logging.getLogger(__name__)


class PlannedTrip(Model):
    __slots__ = [
        'vehicle',
        'trip',
        'down_time',
        'pickup',
        'delivery',
    ]

    vehicle: Vehicle
    trip: Trip
    down_time: float
    pickup: Stop
    delivery: Stop

    def __init__(self, vehicle: Vehicle, trip: Trip, pickup: Stop = None, delivery: Stop = None,
                 down_time: float = 0.0):
        self.vehicle = vehicle
        self.trip = trip
        self.down_time = down_time

        self.pickup = pickup
        self.delivery = delivery

        if self.pickup is not None:
            self.pickup.append_pickup(self)
        if self.delivery is not None:
            self.delivery.append_delivery(self)

    @property
    def pickup_time(self) -> float:
        return self.pickup.arrival_time

    @property
    def delivery_time(self) -> float:
        return self.delivery.arrival_time

    @property
    def trip_identifier(self) -> str:
        return self.trip.identifier

    @property
    def vehicle_identifier(self) -> str:
        return self.vehicle.identifier

    @property
    def origin(self) -> Position:
        return self.trip.origin_position

    @property
    def destination(self) -> Position:
        return self.trip.destination_position

    @property
    def distance(self) -> float:
        return self.trip.distance

    @property
    def duration(self) -> float:
        return self.delivery_time - self.pickup_time

    @property
    def capacity(self):
        return self.trip.capacity

    @cached_property
    def feasible(self) -> bool:
        assert self.pickup in self.delivery.all_previous
        assert self.pickup_time <= self.delivery_time

        if not self.pickup_time <= self.trip.origin_latest:
            return False

        if not self.delivery_time <= self.trip.destination_latest:
            return False

        if not self.pickup.capacity <= self.vehicle.capacity:
            return False

        if not self.duration <= self.trip.timeout:
            return False

        return True

    @property
    def empty(self) -> bool:
        return self.trip.empty

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        yield from (
            ('trip_identifier', self.trip_identifier),
            ('pickup', self.pickup),
            ('delivery', self.delivery),
            ('down_time', self.down_time),
            ('feasible', self.feasible),
        )

    def flush(self) -> None:
        for key in ('feasible',):
            self.__dict__.pop(key, None)
